# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest
from unittest.mock import AsyncMock, MagicMock

from itential_mcp.platform.services.operations_manager import Service
from itential_mcp.core import exceptions
from ipsdk.platform import AsyncPlatform


class TestService:
    """Test the operations_manager Service class"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    def test_service_initialization(self, mock_client):
        """Test Service initialization"""
        service = Service(mock_client)

        assert service.client == mock_client
        assert service.name == "operations_manager"

    def test_service_inheritance(self, service):
        """Test Service inherits from ServiceBase"""
        from itential_mcp.platform.services import ServiceBase

        assert isinstance(service, ServiceBase)
        assert hasattr(service, "client")
        assert hasattr(service, "name")

    def test_service_name_attribute(self, service):
        """Test Service has correct name attribute"""
        assert service.name == "operations_manager"


class TestGetWorkflows:
    """Test the get_workflows method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    @pytest.fixture
    def mock_response_single_page(self):
        """Mock response with workflows that fit in a single page"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": [
                {
                    "_id": "workflow-1",
                    "name": "Test Workflow 1",
                    "type": "endpoint",
                    "enabled": True,
                },
                {
                    "_id": "workflow-2",
                    "name": "Test Workflow 2",
                    "type": "endpoint",
                    "enabled": True,
                },
            ],
            "metadata": {"total": 2, "count": 2, "skip": 0, "limit": 100},
        }
        return mock_response

    @pytest.fixture
    def mock_response_multi_page(self):
        """Mock responses for multi-page workflow results"""
        # First page response - has 1 item but total is 3, so another page needed
        mock_response_1 = MagicMock()
        mock_response_1.json.return_value = {
            "data": [
                {
                    "_id": "workflow-1",
                    "name": "Test Workflow 1",
                    "type": "endpoint",
                    "enabled": True,
                }
            ],
            "metadata": {"total": 3, "count": 1, "skip": 0, "limit": 100},
        }

        # Second page response - has 2 more items, making total received 3
        mock_response_2 = MagicMock()
        mock_response_2.json.return_value = {
            "data": [
                {
                    "_id": "workflow-2",
                    "name": "Test Workflow 2",
                    "type": "endpoint",
                    "enabled": True,
                },
                {
                    "_id": "workflow-3",
                    "name": "Test Workflow 3",
                    "type": "endpoint",
                    "enabled": True,
                },
            ],
            "metadata": {"total": 3, "count": 2, "skip": 100, "limit": 100},
        }

        return [mock_response_1, mock_response_2]

    @pytest.fixture
    def mock_response_empty(self):
        """Mock response with no workflows"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": [],
            "metadata": {"total": 0, "count": 0, "skip": 0, "limit": 100},
        }
        return mock_response

    @pytest.mark.asyncio
    async def test_get_workflows_single_page(self, service, mock_response_single_page):
        """Test get_workflows with single page of results"""
        service.client.get = AsyncMock(return_value=mock_response_single_page)

        result = await service.get_workflows()

        assert len(result) == 2
        assert result[0]["_id"] == "workflow-1"
        assert result[0]["name"] == "Test Workflow 1"
        assert result[1]["_id"] == "workflow-2"
        assert result[1]["name"] == "Test Workflow 2"

        # Verify client.get was called with correct parameters
        service.client.get.assert_called_once_with(
            "/operations-manager/triggers",
            params={
                "limit": 100,
                "skip": 0,
                "equalsField": "type",
                "equals": "endpoint",
                "enabled": True,
            },
        )

    @pytest.mark.asyncio
    async def test_get_workflows_multi_page(self, service, mock_response_multi_page):
        """Test get_workflows with multiple pages of results"""
        service.client.get = AsyncMock(side_effect=mock_response_multi_page)

        result = await service.get_workflows()

        assert len(result) == 3
        assert result[0]["_id"] == "workflow-1"
        assert result[1]["_id"] == "workflow-2"
        assert result[2]["_id"] == "workflow-3"

        # Verify client.get was called twice for pagination
        assert service.client.get.call_count == 2

        # Check first call parameters
        first_call = service.client.get.call_args_list[0]
        assert first_call[0] == ("/operations-manager/triggers",)
        assert first_call[1]["params"]["skip"] == 0

        # Check second call parameters
        second_call = service.client.get.call_args_list[1]
        assert second_call[0] == ("/operations-manager/triggers",)
        assert second_call[1]["params"]["skip"] == 100

    @pytest.mark.asyncio
    async def test_get_workflows_empty_result(self, service, mock_response_empty):
        """Test get_workflows with empty result"""
        service.client.get = AsyncMock(return_value=mock_response_empty)

        result = await service.get_workflows()

        assert len(result) == 0
        assert result == []

        service.client.get.assert_called_once()

    @pytest.mark.asyncio
    async def test_get_workflows_parameters(self, service, mock_response_single_page):
        """Test get_workflows uses correct API parameters"""
        service.client.get = AsyncMock(return_value=mock_response_single_page)

        await service.get_workflows()

        call_args = service.client.get.call_args
        params = call_args[1]["params"]

        assert params["limit"] == 100
        assert params["skip"] == 0
        assert params["equalsField"] == "type"
        assert params["equals"] == "endpoint"
        assert params["enabled"] is True

    @pytest.mark.asyncio
    async def test_get_workflows_client_error(self, service):
        """Test get_workflows handles client errors"""
        service.client.get = AsyncMock(side_effect=Exception("API Error"))

        with pytest.raises(Exception, match="API Error"):
            await service.get_workflows()


class TestStartWorkflow:
    """Test the start_workflow method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    @pytest.fixture
    def mock_workflow_response(self):
        """Mock response for workflow execution"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": {
                "_id": "job-123",
                "name": "Test Workflow",
                "status": "running",
                "tasks": {"task1": {"type": "action"}},
                "metrics": {"start_time": 1640995200000, "user": "user-123"},
            }
        }
        return mock_response

    @pytest.mark.asyncio
    async def test_start_workflow_with_data(self, service, mock_workflow_response):
        """Test start_workflow with data payload"""
        service.client.post = AsyncMock(return_value=mock_workflow_response)

        input_data = {"parameter": "value"}
        result = await service.start_workflow("test-route", input_data)

        expected_result = {
            "_id": "job-123",
            "name": "Test Workflow",
            "status": "running",
            "tasks": {"task1": {"type": "action"}},
            "metrics": {"start_time": 1640995200000, "user": "user-123"},
        }

        assert result == expected_result

        service.client.post.assert_called_once_with(
            "/operations-manager/triggers/endpoint/test-route", json=input_data
        )

    @pytest.mark.asyncio
    async def test_start_workflow_without_data(self, service, mock_workflow_response):
        """Test start_workflow without data payload"""
        service.client.post = AsyncMock(return_value=mock_workflow_response)

        result = await service.start_workflow("test-route")

        assert result["_id"] == "job-123"
        assert result["name"] == "Test Workflow"

        service.client.post.assert_called_once_with(
            "/operations-manager/triggers/endpoint/test-route", json=None
        )

    @pytest.mark.asyncio
    async def test_start_workflow_explicit_none_data(
        self, service, mock_workflow_response
    ):
        """Test start_workflow with explicit None data"""
        service.client.post = AsyncMock(return_value=mock_workflow_response)

        result = await service.start_workflow("test-route", None)

        assert result["_id"] == "job-123"

        service.client.post.assert_called_once_with(
            "/operations-manager/triggers/endpoint/test-route", json=None
        )

    @pytest.mark.asyncio
    async def test_start_workflow_endpoint_path(self, service, mock_workflow_response):
        """Test start_workflow constructs correct endpoint path"""
        service.client.post = AsyncMock(return_value=mock_workflow_response)

        await service.start_workflow("my-custom-route")

        call_args = service.client.post.call_args
        endpoint_path = call_args[0][0]

        assert endpoint_path == "/operations-manager/triggers/endpoint/my-custom-route"

    @pytest.mark.asyncio
    async def test_start_workflow_complex_data(self, service, mock_workflow_response):
        """Test start_workflow with complex data structure"""
        service.client.post = AsyncMock(return_value=mock_workflow_response)

        complex_data = {
            "nested": {"key": "value", "list": [1, 2, 3]},
            "boolean": True,
            "number": 42,
        }

        await service.start_workflow("test-route", complex_data)

        service.client.post.assert_called_once_with(
            "/operations-manager/triggers/endpoint/test-route", json=complex_data
        )

    @pytest.mark.asyncio
    async def test_start_workflow_client_error(self, service):
        """Test start_workflow handles client errors"""
        service.client.post = AsyncMock(side_effect=Exception("API Error"))

        with pytest.raises(Exception, match="API Error"):
            await service.start_workflow("test-route")

    @pytest.mark.asyncio
    async def test_start_workflow_empty_route_name(
        self, service, mock_workflow_response
    ):
        """Test start_workflow with empty route name"""
        service.client.post = AsyncMock(return_value=mock_workflow_response)

        await service.start_workflow("")

        service.client.post.assert_called_once_with(
            "/operations-manager/triggers/endpoint/", json=None
        )


class TestGetJobs:
    """Test the get_jobs method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    @pytest.fixture
    def mock_jobs_response_single_page(self):
        """Mock response with jobs that fit in a single page"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": [
                {
                    "_id": "job-1",
                    "name": "Test Workflow Job 1",
                    "description": "First test job",
                    "status": "complete",
                },
                {
                    "_id": "job-2",
                    "name": "Test Workflow Job 2",
                    "description": "Second test job",
                    "status": "running",
                },
            ],
            "metadata": {"total": 2, "count": 2, "skip": 0, "limit": 100},
        }
        return mock_response

    @pytest.fixture
    def mock_jobs_response_multi_page(self):
        """Mock responses for multi-page job results"""
        # First page response
        mock_response_1 = MagicMock()
        mock_response_1.json.return_value = {
            "data": [
                {
                    "_id": "job-1",
                    "name": "Test Workflow Job 1",
                    "description": "First test job",
                    "status": "complete",
                }
            ],
            "metadata": {"total": 3, "count": 1, "skip": 0, "limit": 100},
        }

        # Second page response
        mock_response_2 = MagicMock()
        mock_response_2.json.return_value = {
            "data": [
                {
                    "_id": "job-2",
                    "name": "Test Workflow Job 2",
                    "description": "Second test job",
                    "status": "running",
                },
                {
                    "_id": "job-3",
                    "name": "Test Workflow Job 3",
                    "description": "Third test job",
                    "status": "error",
                },
            ],
            "metadata": {"total": 3, "count": 2, "skip": 100, "limit": 100},
        }

        return [mock_response_1, mock_response_2]

    @pytest.fixture
    def mock_jobs_response_empty(self):
        """Mock response with no jobs"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": [],
            "metadata": {"total": 0, "count": 0, "skip": 0, "limit": 100},
        }
        return mock_response

    @pytest.mark.asyncio
    async def test_get_jobs_single_page(self, service, mock_jobs_response_single_page):
        """Test get_jobs with single page of results"""
        service.client.get = AsyncMock(return_value=mock_jobs_response_single_page)

        result = await service.get_jobs("Test Workflow")

        assert len(result) == 2
        assert result[0]["_id"] == "job-1"
        assert result[0]["name"] == "Test Workflow Job 1"
        assert result[0]["description"] == "First test job"
        assert result[0]["status"] == "complete"
        assert result[1]["_id"] == "job-2"
        assert result[1]["name"] == "Test Workflow Job 2"
        assert result[1]["status"] == "running"

        # Verify client.get was called with correct parameters
        service.client.get.assert_called_once_with(
            "/operations-manager/jobs",
            params={"limit": 100, "skip": 0, "equals[name]": "Test Workflow"},
        )

    @pytest.mark.asyncio
    async def test_get_jobs_multi_page(self, service, mock_jobs_response_multi_page):
        """Test get_jobs with multiple pages of results"""
        service.client.get = AsyncMock(side_effect=mock_jobs_response_multi_page)

        result = await service.get_jobs("Test Workflow")

        assert len(result) == 3
        assert result[0]["_id"] == "job-1"
        assert result[1]["_id"] == "job-2"
        assert result[2]["_id"] == "job-3"
        assert result[2]["status"] == "error"

        # Verify client.get was called twice for pagination
        assert service.client.get.call_count == 2

        # Check that both calls were made to the correct endpoint
        # Note: params dictionary is reused, so skip values can't be reliably tested
        first_call = service.client.get.call_args_list[0]
        assert first_call[0] == ("/operations-manager/jobs",)
        assert "params" in first_call[1]
        assert first_call[1]["params"]["equals[name]"] == "Test Workflow"
        assert first_call[1]["params"]["limit"] == 100

        second_call = service.client.get.call_args_list[1]
        assert second_call[0] == ("/operations-manager/jobs",)
        assert "params" in second_call[1]
        assert second_call[1]["params"]["equals[name]"] == "Test Workflow"
        assert second_call[1]["params"]["limit"] == 100

    @pytest.mark.asyncio
    async def test_get_jobs_empty_result(self, service, mock_jobs_response_empty):
        """Test get_jobs with empty result"""
        service.client.get = AsyncMock(return_value=mock_jobs_response_empty)

        result = await service.get_jobs("Nonexistent Workflow")

        assert len(result) == 0
        assert result == []

        service.client.get.assert_called_once()

    @pytest.mark.asyncio
    async def test_get_jobs_parameters(self, service, mock_jobs_response_single_page):
        """Test get_jobs uses correct API parameters"""
        service.client.get = AsyncMock(return_value=mock_jobs_response_single_page)

        await service.get_jobs("My Workflow")

        call_args = service.client.get.call_args
        params = call_args[1]["params"]

        assert params["limit"] == 100
        assert params["skip"] == 0
        assert params["equals[name]"] == "My Workflow"
        assert "starts-with[name]" not in params

    @pytest.mark.asyncio
    async def test_get_jobs_with_project_parameter(self, service):
        """Test get_jobs with project parameter (project filtering functionality)"""
        # Mock project lookup response
        mock_project_response = MagicMock()
        mock_project_response.json.return_value = {
            "metadata": {"total": 1},
            "data": [{"_id": "project-123", "name": "test-project"}],
        }

        # Mock jobs response
        mock_jobs_response = MagicMock()
        mock_jobs_response.json.return_value = {
            "data": [
                {
                    "_id": "job-1",
                    "name": "@project-123: Test Workflow",
                    "description": "Project workflow job",
                    "status": "complete",
                }
            ],
            "metadata": {"total": 1},
        }

        service.client.get = AsyncMock(
            side_effect=[mock_project_response, mock_jobs_response]
        )

        result = await service.get_jobs("Test Workflow", "test-project")

        assert len(result) == 1
        assert result[0]["_id"] == "job-1"
        assert result[0]["name"] == "@project-123: Test Workflow"

        # Verify both API calls were made
        assert service.client.get.call_count == 2

        # Check project lookup call
        first_call = service.client.get.call_args_list[0]
        assert first_call[0] == ("/automation-studio/projects",)
        assert first_call[1]["params"]["equals[name]"] == "Test Workflow"

        # Check jobs call with project filtering
        second_call = service.client.get.call_args_list[1]
        assert second_call[0] == ("/operations-manager/jobs",)
        assert second_call[1]["params"]["equals[name]"] == "@project-123: Test Workflow"

    @pytest.mark.asyncio
    async def test_get_jobs_project_not_found(self, service):
        """Test get_jobs raises NotFoundError when project is not found"""
        # Mock empty project response
        mock_project_response = MagicMock()
        mock_project_response.json.return_value = {"metadata": {"total": 0}, "data": []}

        service.client.get = AsyncMock(return_value=mock_project_response)

        with pytest.raises(
            exceptions.NotFoundError,
            match="project nonexistent-project could not be found",
        ):
            await service.get_jobs("Test Workflow", "nonexistent-project")

    @pytest.mark.asyncio
    async def test_get_jobs_with_project_no_workflow_name(self, service):
        """Test get_jobs with project parameter but no specific workflow name"""
        # Mock project lookup response
        mock_project_response = MagicMock()
        mock_project_response.json.return_value = {
            "metadata": {"total": 1},
            "data": [{"_id": "project-456", "name": "test-project"}],
        }

        # Mock jobs response
        mock_jobs_response = MagicMock()
        mock_jobs_response.json.return_value = {"data": [], "metadata": {"total": 0}}

        service.client.get = AsyncMock(
            side_effect=[mock_project_response, mock_jobs_response]
        )

        await service.get_jobs(None, "test-project")

        # Check jobs call uses starts-with filter for all project workflows
        second_call = service.client.get.call_args_list[1]
        assert second_call[1]["params"]["starts-with[name]"] == "@project-456"
        assert "equals[name]" not in second_call[1]["params"]

    @pytest.mark.asyncio
    async def test_get_jobs_client_error(self, service):
        """Test get_jobs handles client errors"""
        service.client.get = AsyncMock(side_effect=Exception("API Error"))

        with pytest.raises(Exception, match="API Error"):
            await service.get_jobs("Test Workflow")

    @pytest.mark.asyncio
    async def test_get_jobs_missing_metadata_fields(self, service):
        """Test get_jobs handles response with missing metadata fields"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": [
                {
                    "_id": "job-1",
                    "name": "Test Job",
                    "description": None,
                    "status": "complete",
                }
            ],
            "metadata": {"total": 1},
        }
        service.client.get = AsyncMock(return_value=mock_response)

        result = await service.get_jobs("Test Workflow")

        assert len(result) == 1
        assert result[0]["_id"] == "job-1"
        assert result[0]["description"] is None

    @pytest.mark.asyncio
    async def test_get_jobs_missing_data_fields(self, service):
        """Test get_jobs handles job data with missing fields"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": [
                {
                    "_id": "job-1",
                    "name": "Test Job",
                    # Missing description and status
                }
            ],
            "metadata": {"total": 1},
        }
        service.client.get = AsyncMock(return_value=mock_response)

        result = await service.get_jobs("Test Workflow")

        assert len(result) == 1
        assert result[0]["_id"] == "job-1"
        assert result[0]["name"] == "Test Job"
        assert result[0]["description"] is None
        assert result[0]["status"] is None


class TestDescribeJob:
    """Test the describe_job method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    @pytest.fixture
    def mock_job_detail_response(self):
        """Mock response for job detail"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": {
                "_id": "job-123",
                "name": "Test Workflow Execution",
                "description": "Detailed test job execution",
                "type": "automation",
                "status": "complete",
                "tasks": {
                    "task1": {
                        "type": "action",
                        "app": "operations-manager",
                        "status": "complete",
                    },
                    "task2": {"type": "manual", "status": "complete"},
                },
                "metrics": {
                    "start_time": 1640995200000,
                    "end_time": 1640995500000,
                    "user": "user-123",
                },
                "last_updated": "2025-01-01T12:05:00Z",
                "created": "2025-01-01T12:00:00Z",
            }
        }
        return mock_response

    @pytest.fixture
    def mock_job_minimal_response(self):
        """Mock response for job with minimal fields"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": {"_id": "job-minimal", "name": "Minimal Job", "status": "running"}
        }
        return mock_response

    @pytest.mark.asyncio
    async def test_describe_job_complete_response(
        self, service, mock_job_detail_response
    ):
        """Test describe_job with complete job details"""
        service.client.get = AsyncMock(return_value=mock_job_detail_response)

        result = await service.describe_job("job-123")

        expected_data = {
            "_id": "job-123",
            "name": "Test Workflow Execution",
            "description": "Detailed test job execution",
            "type": "automation",
            "status": "complete",
            "tasks": {
                "task1": {
                    "type": "action",
                    "app": "operations-manager",
                    "status": "complete",
                },
                "task2": {"type": "manual", "status": "complete"},
            },
            "metrics": {
                "start_time": 1640995200000,
                "end_time": 1640995500000,
                "user": "user-123",
            },
            "last_updated": "2025-01-01T12:05:00Z",
            "created": "2025-01-01T12:00:00Z",
        }

        assert result == expected_data

        service.client.get.assert_called_once_with("/operations-manager/jobs/job-123")

    @pytest.mark.asyncio
    async def test_describe_job_minimal_response(
        self, service, mock_job_minimal_response
    ):
        """Test describe_job with minimal job details"""
        service.client.get = AsyncMock(return_value=mock_job_minimal_response)

        result = await service.describe_job("job-minimal")

        expected_data = {
            "_id": "job-minimal",
            "name": "Minimal Job",
            "status": "running",
        }

        assert result == expected_data

        service.client.get.assert_called_once_with(
            "/operations-manager/jobs/job-minimal"
        )

    @pytest.mark.asyncio
    async def test_describe_job_endpoint_path(self, service, mock_job_detail_response):
        """Test describe_job constructs correct endpoint path"""
        service.client.get = AsyncMock(return_value=mock_job_detail_response)

        await service.describe_job("custom-job-id-123")

        call_args = service.client.get.call_args
        endpoint_path = call_args[0][0]

        assert endpoint_path == "/operations-manager/jobs/custom-job-id-123"

    @pytest.mark.asyncio
    async def test_describe_job_client_error(self, service):
        """Test describe_job handles client errors"""
        service.client.get = AsyncMock(side_effect=Exception("Job not found"))

        with pytest.raises(Exception, match="Job not found"):
            await service.describe_job("nonexistent-job")

    @pytest.mark.asyncio
    async def test_describe_job_empty_object_id(
        self, service, mock_job_detail_response
    ):
        """Test describe_job with empty object ID"""
        service.client.get = AsyncMock(return_value=mock_job_detail_response)

        await service.describe_job("")

        service.client.get.assert_called_once_with("/operations-manager/jobs/")

    @pytest.mark.asyncio
    async def test_describe_job_different_statuses(self, service):
        """Test describe_job with jobs in different statuses"""
        statuses = ["running", "complete", "error", "paused", "canceled", "incomplete"]

        for status in statuses:
            mock_response = MagicMock()
            mock_response.json.return_value = {
                "data": {
                    "_id": f"job-{status}",
                    "name": f"Job {status.title()}",
                    "status": status,
                }
            }

            service.client.get = AsyncMock(return_value=mock_response)

            result = await service.describe_job(f"job-{status}")

            assert result["status"] == status
            assert result["_id"] == f"job-{status}"

    @pytest.mark.asyncio
    async def test_describe_job_complex_tasks(self, service):
        """Test describe_job with complex task structure"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": {
                "_id": "job-complex",
                "name": "Complex Workflow",
                "status": "running",
                "tasks": {
                    "init": {
                        "type": "automation",
                        "app": "operations-manager",
                        "status": "complete",
                        "data": {"initialized": True},
                    },
                    "process": {
                        "type": "automation",
                        "app": "device-manager",
                        "status": "running",
                        "data": {"device_count": 5},
                    },
                    "finalize": {
                        "type": "manual",
                        "status": "pending",
                        "description": "Manual approval required",
                    },
                },
            }
        }

        service.client.get = AsyncMock(return_value=mock_response)

        result = await service.describe_job("job-complex")

        assert "tasks" in result
        assert len(result["tasks"]) == 3
        assert result["tasks"]["init"]["status"] == "complete"
        assert result["tasks"]["process"]["status"] == "running"
        assert result["tasks"]["finalize"]["status"] == "pending"


class TestStartJob:
    """Test the start_job method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    @pytest.fixture
    def mock_start_job_response(self):
        """Mock response for start_job"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "_id": "job-789",
            "name": "Direct Job Execution",
            "description": "Job started directly via workflow name",
            "status": "running",
            "type": "automation",
            "workflow": "test-workflow",
            "variables": {"param1": "value1", "param2": "value2"},
            "created": "2025-01-01T12:00:00Z",
        }
        return mock_response

    @pytest.mark.asyncio
    async def test_start_job_success(self, service, mock_start_job_response):
        """Test start_job with complete parameters"""
        service.client.post = AsyncMock(return_value=mock_start_job_response)

        workflow_name = "test-workflow"
        description = "Test job description"
        variables = {"param1": "value1", "param2": "value2"}

        result = await service.start_job(workflow_name, description, variables)

        expected_result = {
            "_id": "job-789",
            "name": "Direct Job Execution",
            "description": "Job started directly via workflow name",
            "status": "running",
            "type": "automation",
            "workflow": "test-workflow",
            "variables": {"param1": "value1", "param2": "value2"},
            "created": "2025-01-01T12:00:00Z",
        }

        assert result == expected_result

        # Verify API call
        service.client.post.assert_called_once_with(
            "/operations_manager/jobs/start",
            json={
                "workflow": workflow_name,
                "options": {
                    "type": "automation",
                    "groups": [],
                    "description": description,
                    "variables": variables,
                },
            },
        )

    @pytest.mark.asyncio
    async def test_start_job_empty_description(self, service, mock_start_job_response):
        """Test start_job with empty description"""
        service.client.post = AsyncMock(return_value=mock_start_job_response)

        await service.start_job("test-workflow", "", {"key": "value"})

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["options"]["description"] == ""

    @pytest.mark.asyncio
    async def test_start_job_none_description(self, service, mock_start_job_response):
        """Test start_job with None description (should convert to empty string)"""
        service.client.post = AsyncMock(return_value=mock_start_job_response)

        await service.start_job("test-workflow", None, {"key": "value"})

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["options"]["description"] == ""

    @pytest.mark.asyncio
    async def test_start_job_empty_variables(self, service, mock_start_job_response):
        """Test start_job with empty variables dict"""
        service.client.post = AsyncMock(return_value=mock_start_job_response)

        await service.start_job("test-workflow", "description", {})

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["options"]["variables"] == {}

    @pytest.mark.asyncio
    async def test_start_job_none_variables(self, service, mock_start_job_response):
        """Test start_job with None variables (should convert to empty dict)"""
        service.client.post = AsyncMock(return_value=mock_start_job_response)

        await service.start_job("test-workflow", "description", None)

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["options"]["variables"] == {}

    @pytest.mark.asyncio
    async def test_start_job_complex_variables(self, service, mock_start_job_response):
        """Test start_job with complex variables structure"""
        service.client.post = AsyncMock(return_value=mock_start_job_response)

        complex_variables = {
            "nested": {"key": "value", "list": [1, 2, 3]},
            "boolean": True,
            "number": 42,
            "string": "test",
        }

        await service.start_job("complex-workflow", "Complex test", complex_variables)

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["workflow"] == "complex-workflow"
        assert request_body["options"]["variables"] == complex_variables
        assert request_body["options"]["type"] == "automation"
        assert request_body["options"]["groups"] == []

    @pytest.mark.asyncio
    async def test_start_job_endpoint_path(self, service, mock_start_job_response):
        """Test start_job uses correct API endpoint"""
        service.client.post = AsyncMock(return_value=mock_start_job_response)

        await service.start_job("test-workflow", "description", {})

        call_args = service.client.post.call_args
        endpoint_path = call_args[0][0]

        assert endpoint_path == "/operations_manager/jobs/start"

    @pytest.mark.asyncio
    async def test_start_job_request_structure(self, service, mock_start_job_response):
        """Test start_job creates correct request body structure"""
        service.client.post = AsyncMock(return_value=mock_start_job_response)

        workflow = "structure-test-workflow"
        description = "Structure test description"
        variables = {"test": "variable"}

        await service.start_job(workflow, description, variables)

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        # Verify top-level structure
        assert "workflow" in request_body
        assert "options" in request_body
        assert len(request_body) == 2

        # Verify options structure
        options = request_body["options"]
        assert "type" in options
        assert "groups" in options
        assert "description" in options
        assert "variables" in options
        assert len(options) == 4

        # Verify values
        assert request_body["workflow"] == workflow
        assert options["type"] == "automation"
        assert options["groups"] == []
        assert options["description"] == description
        assert options["variables"] == variables

    @pytest.mark.asyncio
    async def test_start_job_client_error(self, service):
        """Test start_job handles client errors"""
        service.client.post = AsyncMock(side_effect=Exception("Workflow not found"))

        with pytest.raises(Exception, match="Workflow not found"):
            await service.start_job("nonexistent-workflow", "description", {})

    @pytest.mark.asyncio
    async def test_start_job_minimal_response(self, service):
        """Test start_job with minimal API response"""
        mock_response = MagicMock()
        mock_response.json.return_value = {"_id": "job-minimal", "status": "queued"}
        service.client.post = AsyncMock(return_value=mock_response)

        result = await service.start_job("test-workflow", "description", {})

        expected_result = {"_id": "job-minimal", "status": "queued"}
        assert result == expected_result

    @pytest.mark.asyncio
    async def test_start_job_workflow_name_types(
        self, service, mock_start_job_response
    ):
        """Test start_job with different workflow name formats"""
        service.client.post = AsyncMock(return_value=mock_start_job_response)

        workflow_names = [
            "simple-workflow",
            "workflow_with_underscores",
            "workflow-with-dashes",
            "workflowCamelCase",
            "Workflow With Spaces",
            "123-numeric-workflow",
        ]

        for workflow_name in workflow_names:
            await service.start_job(workflow_name, "test", {})

            call_args = service.client.post.call_args
            request_body = call_args[1]["json"]
            assert request_body["workflow"] == workflow_name


class TestCreateAutomation:
    """Test the create_automation method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    @pytest.fixture
    def mock_automation_response(self):
        """Mock response for create_automation"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": {
                "_id": "automation-123",
                "name": "Test Automation",
                "componentType": "workflows",
                "componentId": "workflow-456",
                "description": "Test automation description",
            }
        }
        return mock_response

    @pytest.mark.asyncio
    async def test_create_automation_basic(self, service, mock_automation_response):
        """Test create_automation with basic parameters"""
        service.client.post = AsyncMock(return_value=mock_automation_response)

        result = await service.create_automation("Test Automation", "workflows")

        expected_result = {
            "_id": "automation-123",
            "name": "Test Automation",
            "componentType": "workflows",
            "componentId": "workflow-456",
            "description": "Test automation description",
        }

        assert result == expected_result

        # Verify API call
        service.client.post.assert_called_once_with(
            "/operations-manager/automations",
            json={"name": "Test Automation", "componentType": "workflows"},
        )

    @pytest.mark.asyncio
    async def test_create_automation_with_description(
        self, service, mock_automation_response
    ):
        """Test create_automation with description"""
        service.client.post = AsyncMock(return_value=mock_automation_response)

        await service.create_automation(
            "Test Automation", "workflows", description="Custom description"
        )

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["description"] == "Custom description"

    @pytest.mark.asyncio
    async def test_create_automation_with_workflow_component(
        self, service, mock_automation_response
    ):
        """Test create_automation with workflow component lookup"""
        # Mock workflow lookup response
        mock_workflow_response = MagicMock()
        mock_workflow_response.json.return_value = {
            "total": 1,
            "items": [{"_id": "workflow-789", "name": "Target Workflow"}],
        }

        service.client.get = AsyncMock(return_value=mock_workflow_response)
        service.client.post = AsyncMock(return_value=mock_automation_response)

        await service.create_automation(
            "Workflow Automation",
            "workflows",
            component_name="Target Workflow",
            description="Workflow automation",
        )

        # Verify workflow lookup
        service.client.get.assert_called_once_with(
            "/automation-studio/workflows", params={"equals[name]": "Target Workflow"}
        )

        # Verify automation creation with component ID
        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["componentId"] == "workflow-789"

    @pytest.mark.asyncio
    async def test_create_automation_workflow_not_found(self, service):
        """Test create_automation raises NotFoundError when workflow not found"""
        # Mock empty workflow response
        mock_workflow_response = MagicMock()
        mock_workflow_response.json.return_value = {"total": 0, "items": []}

        service.client.get = AsyncMock(return_value=mock_workflow_response)

        with pytest.raises(
            exceptions.NotFoundError, match="workflow Nonexistent Workflow not found"
        ):
            await service.create_automation(
                "Test Automation", "workflows", component_name="Nonexistent Workflow"
            )

    @pytest.mark.asyncio
    async def test_create_automation_multiple_workflows_found(self, service):
        """Test create_automation raises NotFoundError when multiple workflows found"""
        # Mock multiple workflow response
        mock_workflow_response = MagicMock()
        mock_workflow_response.json.return_value = {
            "total": 2,
            "items": [
                {"_id": "workflow-1", "name": "Duplicate Workflow"},
                {"_id": "workflow-2", "name": "Duplicate Workflow"},
            ],
        }

        service.client.get = AsyncMock(return_value=mock_workflow_response)

        with pytest.raises(
            exceptions.NotFoundError, match="workflow Duplicate Workflow not found"
        ):
            await service.create_automation(
                "Test Automation", "workflows", component_name="Duplicate Workflow"
            )

    @pytest.mark.asyncio
    async def test_create_automation_ucm_compliance_plan(
        self, service, mock_automation_response
    ):
        """Test create_automation with UCM compliance plan component type"""
        service.client.post = AsyncMock(return_value=mock_automation_response)

        # UCM compliance plan doesn't have lookup logic implemented yet
        await service.create_automation(
            "Compliance Automation", "ucm_compliance_plan", component_name="Test Plan"
        )

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        # Should not include componentId for compliance plans (not implemented)
        assert "componentId" not in request_body
        assert request_body["componentType"] == "ucm_compliance_plan"

    @pytest.mark.asyncio
    async def test_create_automation_client_error(self, service):
        """Test create_automation handles client errors"""
        service.client.post = AsyncMock(side_effect=Exception("API Error"))

        with pytest.raises(Exception, match="API Error"):
            await service.create_automation("Test", "workflows")


class TestCreateManualTrigger:
    """Test the create_manual_trigger method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    @pytest.fixture
    def mock_trigger_response(self):
        """Mock response for create_manual_trigger"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": {
                "_id": "trigger-123",
                "name": "Manual Test Trigger",
                "type": "manual",
                "actionId": "automation-456",
                "actionType": "automations",
                "enabled": True,
                "description": "Test manual trigger",
            }
        }
        return mock_response

    @pytest.mark.asyncio
    async def test_create_manual_trigger_success(self, service, mock_trigger_response):
        """Test create_manual_trigger with basic parameters"""
        service.client.post = AsyncMock(return_value=mock_trigger_response)

        result = await service.create_manual_trigger(
            "Manual Test Trigger", "automation-456"
        )

        expected_result = {
            "_id": "trigger-123",
            "name": "Manual Test Trigger",
            "type": "manual",
            "actionId": "automation-456",
            "actionType": "automations",
            "enabled": True,
            "description": "Test manual trigger",
        }

        assert result == expected_result

        # Verify API call
        service.client.post.assert_called_once_with(
            "/operations-manager/triggers",
            json={
                "actionId": "automation-456",
                "actionType": "automations",
                "name": "Manual Test Trigger",
                "type": "manual",
                "description": None,
                "enabled": True,
            },
        )

    @pytest.mark.asyncio
    async def test_create_manual_trigger_with_description(
        self, service, mock_trigger_response
    ):
        """Test create_manual_trigger with description"""
        service.client.post = AsyncMock(return_value=mock_trigger_response)

        await service.create_manual_trigger(
            "Described Trigger",
            "automation-789",
            description="Custom trigger description",
        )

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["description"] == "Custom trigger description"
        assert request_body["actionId"] == "automation-789"

    @pytest.mark.asyncio
    async def test_create_manual_trigger_request_structure(
        self, service, mock_trigger_response
    ):
        """Test create_manual_trigger creates correct request structure"""
        service.client.post = AsyncMock(return_value=mock_trigger_response)

        await service.create_manual_trigger("Test Trigger", "automation-123")

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        # Verify all required fields
        assert request_body["actionId"] == "automation-123"
        assert request_body["actionType"] == "automations"
        assert request_body["name"] == "Test Trigger"
        assert request_body["type"] == "manual"
        assert request_body["enabled"] is True
        assert "description" in request_body

    @pytest.mark.asyncio
    async def test_create_manual_trigger_endpoint_path(
        self, service, mock_trigger_response
    ):
        """Test create_manual_trigger uses correct API endpoint"""
        service.client.post = AsyncMock(return_value=mock_trigger_response)

        await service.create_manual_trigger("Test", "automation-id")

        call_args = service.client.post.call_args
        endpoint_path = call_args[0][0]

        assert endpoint_path == "/operations-manager/triggers"

    @pytest.mark.asyncio
    async def test_create_manual_trigger_client_error(self, service):
        """Test create_manual_trigger handles client errors"""
        service.client.post = AsyncMock(side_effect=Exception("Automation not found"))

        with pytest.raises(Exception, match="Automation not found"):
            await service.create_manual_trigger("Test", "nonexistent-automation")


class TestCreateEndpointTrigger:
    """Test the create_endpoint_trigger method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    @pytest.fixture
    def mock_endpoint_trigger_response(self):
        """Mock response for create_endpoint_trigger"""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": {
                "_id": "trigger-456",
                "name": "API Endpoint Trigger",
                "type": "endpoint",
                "verb": "POST",
                "routeName": "test-route",
                "actionId": "automation-789",
                "actionType": "automations",
                "enabled": True,
                "schema": {"type": "object", "properties": {}},
            }
        }
        return mock_response

    @pytest.mark.asyncio
    async def test_create_endpoint_trigger_basic(
        self, service, mock_endpoint_trigger_response
    ):
        """Test create_endpoint_trigger with basic parameters"""
        service.client.post = AsyncMock(return_value=mock_endpoint_trigger_response)

        result = await service.create_endpoint_trigger(
            "API Endpoint Trigger", "automation-789", "test-route"
        )

        expected_result = {
            "_id": "trigger-456",
            "name": "API Endpoint Trigger",
            "type": "endpoint",
            "verb": "POST",
            "routeName": "test-route",
            "actionId": "automation-789",
            "actionType": "automations",
            "enabled": True,
            "schema": {"type": "object", "properties": {}},
        }

        assert result == expected_result

        # Verify API call with default schema
        service.client.post.assert_called_once_with(
            "/operations-manager/triggers",
            json={
                "actionId": "automation-789",
                "actionType": "automations",
                "name": "API Endpoint Trigger",
                "type": "endpoint",
                "verb": "POST",
                "routeName": "test-route",
                "description": None,
                "enabled": True,
                "schema": {
                    "type": "object",
                    "properties": {},
                    "additionalProperties": True,
                },
            },
        )

    @pytest.mark.asyncio
    async def test_create_endpoint_trigger_with_custom_schema(
        self, service, mock_endpoint_trigger_response
    ):
        """Test create_endpoint_trigger with custom schema"""
        service.client.post = AsyncMock(return_value=mock_endpoint_trigger_response)

        custom_schema = {
            "type": "object",
            "properties": {"device": {"type": "string"}, "action": {"type": "string"}},
            "required": ["device"],
        }

        await service.create_endpoint_trigger(
            "Custom Schema Trigger",
            "automation-123",
            "custom-route",
            schema=custom_schema,
        )

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["schema"] == custom_schema

    @pytest.mark.asyncio
    async def test_create_endpoint_trigger_with_description(
        self, service, mock_endpoint_trigger_response
    ):
        """Test create_endpoint_trigger with description"""
        service.client.post = AsyncMock(return_value=mock_endpoint_trigger_response)

        await service.create_endpoint_trigger(
            "Described Trigger",
            "automation-456",
            "described-route",
            description="Custom endpoint description",
        )

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        assert request_body["description"] == "Custom endpoint description"

    @pytest.mark.asyncio
    async def test_create_endpoint_trigger_route_validation(self, service):
        """Test create_endpoint_trigger validates route name"""
        # Mock stringutils.is_valid_url_path to return False
        with pytest.raises(ValueError, match="route_name is invalid"):
            await service.create_endpoint_trigger(
                "Invalid Route Trigger",
                "automation-123",
                "invalid route with spaces",  # Invalid route name
            )

    @pytest.mark.asyncio
    async def test_create_endpoint_trigger_valid_route_names(
        self, service, mock_endpoint_trigger_response
    ):
        """Test create_endpoint_trigger with various valid route names"""
        service.client.post = AsyncMock(return_value=mock_endpoint_trigger_response)

        valid_routes = [
            "simple-route",
            "route_with_underscores",
            "route123",
            "route/with/slashes",
            "complex-route_123/with.dots",
        ]

        for route in valid_routes:
            await service.create_endpoint_trigger(
                f"Trigger for {route}", "automation-test", route
            )

            call_args = service.client.post.call_args
            request_body = call_args[1]["json"]
            assert request_body["routeName"] == route

    @pytest.mark.asyncio
    async def test_create_endpoint_trigger_request_structure(
        self, service, mock_endpoint_trigger_response
    ):
        """Test create_endpoint_trigger creates correct request structure"""
        service.client.post = AsyncMock(return_value=mock_endpoint_trigger_response)

        await service.create_endpoint_trigger(
            "Structure Test Trigger", "automation-structure", "structure-route"
        )

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        # Verify all required fields
        assert request_body["actionId"] == "automation-structure"
        assert request_body["actionType"] == "automations"
        assert request_body["name"] == "Structure Test Trigger"
        assert request_body["type"] == "endpoint"
        assert request_body["verb"] == "POST"
        assert request_body["routeName"] == "structure-route"
        assert request_body["enabled"] is True
        assert "schema" in request_body

    @pytest.mark.asyncio
    async def test_create_endpoint_trigger_default_schema_structure(
        self, service, mock_endpoint_trigger_response
    ):
        """Test create_endpoint_trigger default schema structure"""
        service.client.post = AsyncMock(return_value=mock_endpoint_trigger_response)

        await service.create_endpoint_trigger(
            "Default Schema Test", "automation-default", "default-route"
        )

        call_args = service.client.post.call_args
        request_body = call_args[1]["json"]

        expected_schema = {
            "type": "object",
            "properties": {},
            "additionalProperties": True,
        }

        assert request_body["schema"] == expected_schema

    @pytest.mark.asyncio
    async def test_create_endpoint_trigger_client_error(self, service):
        """Test create_endpoint_trigger handles client errors"""
        service.client.post = AsyncMock(side_effect=Exception("Route already exists"))

        with pytest.raises(Exception, match="Route already exists"):
            await service.create_endpoint_trigger(
                "Error Test", "automation-error", "error-route"
            )


class TestDeleteAutomation:
    """Test the delete_automation method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock AsyncPlatform client"""
        return AsyncMock(spec=AsyncPlatform)

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mock client"""
        return Service(mock_client)

    @pytest.fixture
    def mock_delete_response(self):
        """Mock response for delete_automation"""
        mock_response = MagicMock()
        mock_response.json.return_value = {"message": "Automation successfully deleted"}
        return mock_response

    @pytest.mark.asyncio
    async def test_delete_automation_success(self, service, mock_delete_response):
        """Test delete_automation successful deletion"""
        service.client.delete = AsyncMock(return_value=mock_delete_response)

        result = await service.delete_automation("automation-123")

        expected_result = {"message": "Automation successfully deleted"}

        assert result == expected_result

        # Verify API call
        service.client.delete.assert_called_once_with(
            "/operations-manager/automations/automation-123"
        )

    @pytest.mark.asyncio
    async def test_delete_automation_endpoint_path(self, service, mock_delete_response):
        """Test delete_automation constructs correct endpoint path"""
        service.client.delete = AsyncMock(return_value=mock_delete_response)

        await service.delete_automation("automation-custom-id-456")

        call_args = service.client.delete.call_args
        endpoint_path = call_args[0][0]

        assert (
            endpoint_path == "/operations-manager/automations/automation-custom-id-456"
        )

    @pytest.mark.asyncio
    async def test_delete_automation_client_error(self, service):
        """Test delete_automation handles client errors"""
        service.client.delete = AsyncMock(side_effect=Exception("Automation not found"))

        with pytest.raises(Exception, match="Automation not found"):
            await service.delete_automation("nonexistent-automation")

    @pytest.mark.asyncio
    async def test_delete_automation_different_message_formats(self, service):
        """Test delete_automation handles different response message formats"""
        messages = [
            "Automation deleted successfully",
            "Successfully removed automation",
            "Automation removal completed",
        ]

        for message in messages:
            mock_response = MagicMock()
            mock_response.json.return_value = {"message": message}
            service.client.delete = AsyncMock(return_value=mock_response)

            result = await service.delete_automation(f"automation-{message[:5]}")

            assert result["message"] == message

    @pytest.mark.asyncio
    async def test_delete_automation_empty_id(self, service, mock_delete_response):
        """Test delete_automation with empty automation ID"""
        service.client.delete = AsyncMock(return_value=mock_delete_response)

        await service.delete_automation("")

        # Should still make the API call with empty ID
        service.client.delete.assert_called_once_with(
            "/operations-manager/automations/"
        )



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest
import asyncio
from unittest.mock import patch
from io import StringIO
import sys

from itential_mcp.core import metadata


class TestMetadataConstants:
    """Test cases for metadata constants."""

    def test_name_constant_exists(self):
        """Test that name constant exists and has correct value."""
        assert hasattr(metadata, "name")
        assert metadata.name == "itential-mcp"
        assert isinstance(metadata.name, str)

    def test_author_constant_exists(self):
        """Test that author constant exists and has correct value."""
        assert hasattr(metadata, "author")
        assert metadata.author == "Itential"
        assert isinstance(metadata.author, str)

    def test_version_constant_exists(self):
        """Test that version constant exists and is a string."""
        assert hasattr(metadata, "version")
        assert isinstance(metadata.version, str)
        assert len(metadata.version) > 0

    def test_constants_are_immutable_types(self):
        """Test that constants are immutable types."""
        assert isinstance(metadata.name, str)
        assert isinstance(metadata.author, str)
        assert isinstance(metadata.version, str)


class TestDisplayVersionFunction:
    """Test cases for the display_version function."""

    def test_display_version_function_exists(self):
        """Test that display_version function exists and is callable."""
        assert hasattr(metadata, "display_version")
        assert callable(metadata.display_version)

    def test_display_version_is_async(self):
        """Test that display_version is an async function."""
        assert asyncio.iscoroutinefunction(metadata.display_version)

    def test_display_version_function_signature(self):
        """Test that display_version has correct signature."""
        import inspect

        sig = inspect.signature(metadata.display_version)
        params = list(sig.parameters.keys())

        # Should have no parameters
        assert len(params) == 0

        # Should return None
        assert sig.return_annotation is None or sig.return_annotation is type(None)

    def test_display_version_docstring(self):
        """Test that display_version has proper docstring."""
        assert metadata.display_version.__doc__ is not None
        assert (
            "Prints the current app verion to stdout"
            in metadata.display_version.__doc__
        )
        assert (
            "print the current application version to stdout"
            in metadata.display_version.__doc__
        )

    @pytest.mark.asyncio
    @patch("builtins.print")
    async def test_display_version_prints_correct_format(self, mock_print):
        """Test that display_version prints in correct format."""
        await metadata.display_version()

        # Should have been called once
        mock_print.assert_called_once()

        # Get the actual call arguments
        call_args = mock_print.call_args[0][0]

        # Should contain name and version with newline
        assert metadata.name in call_args
        assert metadata.version in call_args
        assert call_args.endswith("\n")
        assert " " in call_args  # Should have space between name and version

    @pytest.mark.asyncio
    @patch("builtins.print")
    async def test_display_version_uses_module_variables(self, mock_print):
        """Test that display_version uses the module's name and version variables."""
        await metadata.display_version()

        call_args = mock_print.call_args[0][0]
        expected_output = f"{metadata.name} {metadata.version}\n"

        assert call_args == expected_output

    @pytest.mark.asyncio
    @patch("builtins.print")
    async def test_display_version_multiple_calls(self, mock_print):
        """Test that display_version can be called multiple times."""
        await metadata.display_version()
        await metadata.display_version()
        await metadata.display_version()

        # Should have been called three times
        assert mock_print.call_count == 3

        # All calls should be identical
        calls = mock_print.call_args_list
        for call in calls:
            assert call[0][0] == f"{metadata.name} {metadata.version}\n"

    @pytest.mark.asyncio
    @patch("sys.stdout", new_callable=StringIO)
    async def test_display_version_real_output(self, mock_stdout):
        """Test display_version with real stdout capture."""
        await metadata.display_version()

        output = mock_stdout.getvalue()
        assert f"{metadata.name} {metadata.version}\n\n" == output

    @pytest.mark.asyncio
    @patch("builtins.print")
    async def test_display_version_returns_none(self, mock_print):
        """Test that display_version returns None."""
        result = await metadata.display_version()
        assert result is None

    @pytest.mark.asyncio
    @patch("builtins.print", side_effect=Exception("Print error"))
    async def test_display_version_print_exception(self, mock_print):
        """Test display_version behavior when print raises exception."""
        with pytest.raises(Exception, match="Print error"):
            await metadata.display_version()


class TestMetadataModule:
    """Test cases for the metadata module as a whole."""

    def test_module_imports(self):
        """Test that the module has correct imports."""
        assert hasattr(metadata, "version")
        # Note: the module version variable shadows the imported version function

    def test_module_structure(self):
        """Test the overall module structure."""
        # Check that expected attributes are present
        expected_attrs = ["name", "author", "version", "display_version"]

        # Get all public attributes (not starting with _)
        public_attrs = [attr for attr in dir(metadata) if not attr.startswith("_")]

        # Should contain at least our expected attributes
        for attr in expected_attrs:
            assert attr in public_attrs

    def test_module_has_copyright(self):
        """Test that the module file has copyright header."""
        import itential_mcp.core.metadata
        import inspect

        source = inspect.getsource(itential_mcp.core.metadata)
        assert "Copyright (c) 2025 Itential, Inc" in source
        assert "GNU General Public License v3.0+" in source

    def test_version_initialization(self):
        """Test that version is properly initialized from importlib.metadata."""
        # Just verify the version is a string and not empty
        assert isinstance(metadata.version, str)
        assert len(metadata.version) > 0


class TestMetadataIntegration:
    """Integration tests for metadata functionality."""

    def test_version_is_valid_semantic_version(self):
        """Test that version follows semantic versioning pattern."""
        version_str = metadata.version

        # Should be a non-empty string
        assert isinstance(version_str, str)
        assert len(version_str) > 0

        # Should not contain newlines or control characters
        assert "\n" not in version_str
        assert "\r" not in version_str
        assert "\t" not in version_str

    def test_constants_consistency(self):
        """Test that constants are consistent across module."""
        # Name should be consistent with package name
        assert metadata.name == "itential-mcp"

        # Author should be consistent
        assert metadata.author == "Itential"

        # Version should be accessible and consistent
        version1 = metadata.version
        version2 = metadata.version
        assert version1 == version2

    @pytest.mark.asyncio
    async def test_display_version_integration(self):
        """Integration test for display_version function."""
        # Capture stdout
        old_stdout = sys.stdout
        sys.stdout = captured_output = StringIO()

        try:
            await metadata.display_version()
            output = captured_output.getvalue()

            # Should contain expected format
            assert metadata.name in output
            assert metadata.version in output
            assert output.endswith("\n")

            # Should be in format "name version\n\n" (function adds extra newline)
            expected = f"{metadata.name} {metadata.version}\n\n"
            assert output == expected

        finally:
            sys.stdout = old_stdout

    def test_module_can_be_imported_multiple_times(self):
        """Test that module can be imported multiple times safely."""
        import itential_mcp.core.metadata as meta1
        import itential_mcp.core.metadata as meta2
        from itential_mcp.core import metadata as meta3

        # All should reference the same module
        assert meta1 is meta2 is meta3

        # Constants should be the same
        assert meta1.name == meta2.name == meta3.name
        assert meta1.author == meta2.author == meta3.author
        assert meta1.version == meta2.version == meta3.version

    @pytest.mark.asyncio
    async def test_display_version_concurrent_calls(self):
        """Test that display_version handles concurrent calls safely."""
        import asyncio

        # Capture stdout
        old_stdout = sys.stdout
        sys.stdout = captured_output = StringIO()

        try:
            # Run multiple concurrent calls
            tasks = [metadata.display_version() for _ in range(5)]
            await asyncio.gather(*tasks)

            output = captured_output.getvalue()

            # Should have 5 lines of output
            lines = output.split("\n")
            assert len([line for line in lines if line]) == 5

            # Each pair of lines should be the same (function outputs extra newline)
            expected_line = f"{metadata.name} {metadata.version}"
            non_empty_lines = [line for line in lines if line]
            assert len(non_empty_lines) == 5
            for line in non_empty_lines:
                assert line == expected_line

        finally:
            sys.stdout = old_stdout


class TestMetadataErrorHandling:
    """Test error handling scenarios."""

    def test_version_import_error_handling(self):
        """Test that version is properly loaded from package metadata."""
        # Since module is already imported, just verify version exists
        assert hasattr(metadata, "version")
        assert metadata.version is not None

    @pytest.mark.asyncio
    @patch("builtins.print")
    async def test_display_version_no_stdout(self, mock_print):
        """Test display_version behavior when print is mocked."""
        await metadata.display_version()
        mock_print.assert_called_once_with(f"{metadata.name} {metadata.version}\n")

    def test_constants_are_not_none(self):
        """Test that constants are not None."""
        assert metadata.name is not None
        assert metadata.author is not None
        assert metadata.version is not None

    def test_constants_are_strings(self):
        """Test that all constants are strings."""
        assert isinstance(metadata.name, str)
        assert isinstance(metadata.author, str)
        # Version should be a string, not a Mock object
        assert isinstance(metadata.version, str)



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

"""Constants used throughout the Itential MCP application."""

from typing import Dict, Any
from dataclasses import dataclass

# Application constants
APP_NAME = "itential-mcp"
APP_DESCRIPTION = "Itential MCP\n\n  Find more information at: https://github.com/itential/itential-mcp"
GITHUB_URL = "https://github.com/itential/itential-mcp"

# Environment variable constants
ENV_PREFIX = "ITENTIAL_MCP_"
CONFIG_ENV_VAR = "ITENTIAL_MCP_CONFIG"

# Help messages
GLOBAL_HELP_MESSAGE = "Prints this help message and exits"
CONFIG_HELP_MESSAGE = "The Itential MCP configuration file"
COMMAND_HELP_SUFFIX = (
    '\nUse "itential-mcp <COMMAND> --help" for more information about a command.\n'
)

# Error messages
HANDLER_TYPE_ERROR = "handler must be callable and awaitable"


@dataclass(frozen=True)
class CommandConfig:
    """Configuration for a CLI command."""

    name: str
    description: str
    arguments: Dict[str, Any]
    add_platform_group: bool = False
    add_server_group: bool = False


# Command configurations
COMMANDS = [
    CommandConfig(
        name="run",
        description="Run the MCP server",
        arguments={"--config": {"help": CONFIG_HELP_MESSAGE}},
        add_platform_group=True,
        add_server_group=True,
    ),
    CommandConfig(
        name="call",
        description="Call a tool and return the results",
        arguments={
            "tool": {"help": "Name of the tool call"},
            "--params": {
                "metavar": "<object>",
                "help": "Parameters to pass to the tool",
            },
        },
        add_platform_group=True,
    ),
    CommandConfig(
        name="tools",
        description="Get list of available tools",
        arguments={},
    ),
    CommandConfig(
        name="tags",
        description="Get list of available tags",
        arguments={},
    ),
    CommandConfig(
        name="version",
        description="Print the version information",
        arguments={},
    ),
]



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest
from pydantic import ValidationError

from itential_mcp.models.configuration_manager import (
    GoldenConfigTree,
    GetGoldenConfigTreesResponse,
    CreateGoldenConfigTreeResponse,
    AddGoldenConfigNodeResponse,
    RenderTemplateResponse,
)


class TestGoldenConfigTree:
    """Test cases for GoldenConfigTree model."""

    def test_create_valid_golden_config_tree(self):
        """Test creating a valid GoldenConfigTree instance."""
        tree_data = {
            "name": "test-tree",
            "device_type": "cisco_ios",
            "versions": ["v1.0", "v2.0", "v3.0"],
        }

        tree = GoldenConfigTree(**tree_data)

        assert tree.name == "test-tree"
        assert tree.device_type == "cisco_ios"
        assert tree.versions == ["v1.0", "v2.0", "v3.0"]

    def test_create_golden_config_tree_with_empty_versions(self):
        """Test creating a GoldenConfigTree with empty versions list."""
        tree_data = {
            "name": "empty-versions-tree",
            "device_type": "juniper",
            "versions": [],
        }

        tree = GoldenConfigTree(**tree_data)

        assert tree.name == "empty-versions-tree"
        assert tree.device_type == "juniper"
        assert tree.versions == []

    def test_create_golden_config_tree_with_single_version(self):
        """Test creating a GoldenConfigTree with a single version."""
        tree_data = {
            "name": "single-version-tree",
            "device_type": "arista_eos",
            "versions": ["initial"],
        }

        tree = GoldenConfigTree(**tree_data)

        assert tree.name == "single-version-tree"
        assert tree.device_type == "arista_eos"
        assert tree.versions == ["initial"]

    def test_golden_config_tree_missing_name(self):
        """Test GoldenConfigTree validation with missing name field."""
        tree_data = {"device_type": "cisco_ios", "versions": ["v1.0"]}

        with pytest.raises(ValidationError) as exc_info:
            GoldenConfigTree(**tree_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "missing"
        assert errors[0]["loc"] == ("name",)

    def test_golden_config_tree_missing_device_type(self):
        """Test GoldenConfigTree validation with missing device_type field."""
        tree_data = {"name": "test-tree", "versions": ["v1.0"]}

        with pytest.raises(ValidationError) as exc_info:
            GoldenConfigTree(**tree_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "missing"
        assert errors[0]["loc"] == ("device_type",)

    def test_golden_config_tree_missing_versions(self):
        """Test GoldenConfigTree validation with missing versions field."""
        tree_data = {"name": "test-tree", "device_type": "cisco_ios"}

        with pytest.raises(ValidationError) as exc_info:
            GoldenConfigTree(**tree_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "missing"
        assert errors[0]["loc"] == ("versions",)

    def test_golden_config_tree_invalid_name_type(self):
        """Test GoldenConfigTree validation with invalid name type."""
        tree_data = {"name": 123, "device_type": "cisco_ios", "versions": ["v1.0"]}

        with pytest.raises(ValidationError) as exc_info:
            GoldenConfigTree(**tree_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "string_type"
        assert errors[0]["loc"] == ("name",)

    def test_golden_config_tree_invalid_device_type(self):
        """Test GoldenConfigTree validation with invalid device_type type."""
        tree_data = {"name": "test-tree", "device_type": None, "versions": ["v1.0"]}

        with pytest.raises(ValidationError) as exc_info:
            GoldenConfigTree(**tree_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "string_type"
        assert errors[0]["loc"] == ("device_type",)

    def test_golden_config_tree_invalid_versions_type(self):
        """Test GoldenConfigTree validation with invalid versions type."""
        tree_data = {
            "name": "test-tree",
            "device_type": "cisco_ios",
            "versions": "not-a-list",
        }

        with pytest.raises(ValidationError) as exc_info:
            GoldenConfigTree(**tree_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "list_type"
        assert errors[0]["loc"] == ("versions",)

    def test_golden_config_tree_invalid_version_item_type(self):
        """Test GoldenConfigTree validation with invalid version item type."""
        tree_data = {
            "name": "test-tree",
            "device_type": "cisco_ios",
            "versions": ["v1.0", 123, "v3.0"],
        }

        with pytest.raises(ValidationError) as exc_info:
            GoldenConfigTree(**tree_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "string_type"
        assert errors[0]["loc"] == ("versions", 1)

    def test_golden_config_tree_serialization(self):
        """Test GoldenConfigTree model serialization to dict."""
        tree_data = {
            "name": "serialization-tree",
            "device_type": "cisco_nxos",
            "versions": ["v1.0", "v2.0"],
        }

        tree = GoldenConfigTree(**tree_data)
        serialized = tree.model_dump()

        assert serialized == tree_data

    def test_golden_config_tree_json_serialization(self):
        """Test GoldenConfigTree model JSON serialization."""
        tree_data = {
            "name": "json-tree",
            "device_type": "juniper_junos",
            "versions": ["initial", "v2.1"],
        }

        tree = GoldenConfigTree(**tree_data)
        json_str = tree.model_dump_json()

        assert '"name":"json-tree"' in json_str
        assert '"device_type":"juniper_junos"' in json_str
        assert '"versions":["initial","v2.1"]' in json_str


class TestGetGoldenConfigTreesResponse:
    """Test cases for GetGoldenConfigTreesResponse model."""

    def test_create_valid_response_with_multiple_trees(self):
        """Test creating a valid GetGoldenConfigTreesResponse with multiple trees."""
        trees_data = [
            {"name": "tree1", "device_type": "cisco_ios", "versions": ["v1.0", "v2.0"]},
            {"name": "tree2", "device_type": "juniper", "versions": ["initial"]},
        ]

        response = GetGoldenConfigTreesResponse(root=trees_data)

        assert len(response.root) == 2
        assert response.root[0].name == "tree1"
        assert response.root[0].device_type == "cisco_ios"
        assert response.root[0].versions == ["v1.0", "v2.0"]
        assert response.root[1].name == "tree2"
        assert response.root[1].device_type == "juniper"
        assert response.root[1].versions == ["initial"]

    def test_create_response_with_empty_list(self):
        """Test creating GetGoldenConfigTreesResponse with empty list."""
        response = GetGoldenConfigTreesResponse(root=[])

        assert response.root == []
        assert len(response.root) == 0

    def test_create_response_with_single_tree(self):
        """Test creating GetGoldenConfigTreesResponse with single tree."""
        tree_data = {
            "name": "single-tree",
            "device_type": "arista_eos",
            "versions": ["v1.0", "v1.1", "v2.0"],
        }

        response = GetGoldenConfigTreesResponse(root=[tree_data])

        assert len(response.root) == 1
        assert response.root[0].name == "single-tree"
        assert response.root[0].device_type == "arista_eos"
        assert response.root[0].versions == ["v1.0", "v1.1", "v2.0"]

    def test_response_with_invalid_tree_data(self):
        """Test GetGoldenConfigTreesResponse validation with invalid tree data."""
        invalid_trees_data = [
            {"name": "valid-tree", "device_type": "cisco_ios", "versions": ["v1.0"]},
            {"device_type": "juniper"},
        ]

        with pytest.raises(ValidationError) as exc_info:
            GetGoldenConfigTreesResponse(root=invalid_trees_data)

        errors = exc_info.value.errors()
        assert len(errors) >= 1
        missing_field_errors = [e for e in errors if e["type"] == "missing"]
        assert len(missing_field_errors) >= 1

    def test_response_direct_list_initialization(self):
        """Test GetGoldenConfigTreesResponse direct initialization with list."""
        trees_data = [
            {"name": "direct-tree1", "device_type": "cisco_ios", "versions": ["v1.0"]},
            {
                "name": "direct-tree2",
                "device_type": "juniper",
                "versions": ["initial", "v2.0"],
            },
        ]

        response = GetGoldenConfigTreesResponse(trees_data)

        assert len(response.root) == 2
        assert response.root[0].name == "direct-tree1"
        assert response.root[1].name == "direct-tree2"

    def test_response_serialization(self):
        """Test GetGoldenConfigTreesResponse serialization."""
        trees_data = [
            {
                "name": "serialize-tree",
                "device_type": "cisco_nxos",
                "versions": ["v1.0", "v2.0"],
            }
        ]

        response = GetGoldenConfigTreesResponse(root=trees_data)
        serialized = response.model_dump()

        assert serialized == trees_data

    def test_response_iteration(self):
        """Test that GetGoldenConfigTreesResponse root can be iterated."""
        trees_data = [
            {"name": f"tree{i}", "device_type": "cisco_ios", "versions": [f"v{i}.0"]}
            for i in range(3)
        ]

        response = GetGoldenConfigTreesResponse(root=trees_data)

        tree_names = []
        for tree in response.root:
            tree_names.append(tree.name)

        assert tree_names == ["tree0", "tree1", "tree2"]


class TestCreateGoldenConfigTreeResponse:
    """Test cases for CreateGoldenConfigTreeResponse model."""

    def test_create_valid_response(self):
        """Test creating a valid CreateGoldenConfigTreeResponse."""
        response_data = {"name": "created-tree", "device_type": "cisco_ios"}

        response = CreateGoldenConfigTreeResponse(**response_data)

        assert response.name == "created-tree"
        assert response.device_type == "cisco_ios"

    def test_create_response_with_different_device_types(self):
        """Test creating response with various device types."""
        device_types = [
            "cisco_ios",
            "cisco_nxos",
            "juniper_junos",
            "arista_eos",
            "custom_device",
        ]

        for device_type in device_types:
            response_data = {"name": f"tree-{device_type}", "device_type": device_type}

            response = CreateGoldenConfigTreeResponse(**response_data)

            assert response.name == f"tree-{device_type}"
            assert response.device_type == device_type

    def test_response_missing_name(self):
        """Test CreateGoldenConfigTreeResponse validation with missing name."""
        response_data = {"device_type": "cisco_ios"}

        with pytest.raises(ValidationError) as exc_info:
            CreateGoldenConfigTreeResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "missing"
        assert errors[0]["loc"] == ("name",)

    def test_response_missing_device_type(self):
        """Test CreateGoldenConfigTreeResponse validation with missing device_type."""
        response_data = {"name": "test-tree"}

        with pytest.raises(ValidationError) as exc_info:
            CreateGoldenConfigTreeResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "missing"
        assert errors[0]["loc"] == ("device_type",)

    def test_response_invalid_name_type(self):
        """Test CreateGoldenConfigTreeResponse validation with invalid name type."""
        response_data = {"name": None, "device_type": "cisco_ios"}

        with pytest.raises(ValidationError) as exc_info:
            CreateGoldenConfigTreeResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "string_type"
        assert errors[0]["loc"] == ("name",)

    def test_response_invalid_device_type_type(self):
        """Test CreateGoldenConfigTreeResponse validation with invalid device_type type."""
        response_data = {"name": "test-tree", "device_type": 123}

        with pytest.raises(ValidationError) as exc_info:
            CreateGoldenConfigTreeResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "string_type"
        assert errors[0]["loc"] == ("device_type",)

    def test_response_serialization(self):
        """Test CreateGoldenConfigTreeResponse serialization."""
        response_data = {"name": "serialize-tree", "device_type": "juniper_junos"}

        response = CreateGoldenConfigTreeResponse(**response_data)
        serialized = response.model_dump()

        assert serialized == response_data

    def test_response_json_serialization(self):
        """Test CreateGoldenConfigTreeResponse JSON serialization."""
        response_data = {"name": "json-tree", "device_type": "arista_eos"}

        response = CreateGoldenConfigTreeResponse(**response_data)
        json_str = response.model_dump_json()

        assert '"name":"json-tree"' in json_str
        assert '"device_type":"arista_eos"' in json_str


class TestAddGoldenConfigNodeResponse:
    """Test cases for AddGoldenConfigNodeResponse model."""

    def test_create_valid_response(self):
        """Test creating a valid AddGoldenConfigNodeResponse."""
        response_data = {
            "message": "Node added successfully to the Golden Configuration tree"
        }

        response = AddGoldenConfigNodeResponse(**response_data)

        assert (
            response.message
            == "Node added successfully to the Golden Configuration tree"
        )

    def test_create_response_with_various_messages(self):
        """Test creating response with various success messages."""
        messages = [
            "Node 'interface-config' added to tree 'cisco-base'",
            "Successfully added node to path 'base/routing/ospf'",
            "Node creation completed",
            "Configuration node added with template",
        ]

        for message in messages:
            response_data = {"message": message}
            response = AddGoldenConfigNodeResponse(**response_data)

            assert response.message == message

    def test_response_missing_message(self):
        """Test AddGoldenConfigNodeResponse validation with missing message."""
        response_data = {}

        with pytest.raises(ValidationError) as exc_info:
            AddGoldenConfigNodeResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "missing"
        assert errors[0]["loc"] == ("message",)

    def test_response_invalid_message_type(self):
        """Test AddGoldenConfigNodeResponse validation with invalid message type."""
        response_data = {"message": 12345}

        with pytest.raises(ValidationError) as exc_info:
            AddGoldenConfigNodeResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "string_type"
        assert errors[0]["loc"] == ("message",)

    def test_response_none_message(self):
        """Test AddGoldenConfigNodeResponse validation with None message."""
        response_data = {"message": None}

        with pytest.raises(ValidationError) as exc_info:
            AddGoldenConfigNodeResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "string_type"
        assert errors[0]["loc"] == ("message",)

    def test_response_empty_message(self):
        """Test AddGoldenConfigNodeResponse with empty message string."""
        response_data = {"message": ""}

        response = AddGoldenConfigNodeResponse(**response_data)

        assert response.message == ""

    def test_response_serialization(self):
        """Test AddGoldenConfigNodeResponse serialization."""
        response_data = {"message": "Node successfully added to configuration tree"}

        response = AddGoldenConfigNodeResponse(**response_data)
        serialized = response.model_dump()

        assert serialized == response_data

    def test_response_json_serialization(self):
        """Test AddGoldenConfigNodeResponse JSON serialization."""
        response_data = {"message": "Configuration node added successfully"}

        response = AddGoldenConfigNodeResponse(**response_data)
        json_str = response.model_dump_json()

        assert '"message":"Configuration node added successfully"' in json_str

    def test_response_with_special_characters(self):
        """Test AddGoldenConfigNodeResponse with special characters in message."""
        special_messages = [
            "Node 'test-node' added with template 'base/config'",
            "Success: Added node @path='/base/interfaces'",
            "Node creation completed with 100% success rate",
            "Added node with config: {'key': 'value'}",
        ]

        for message in special_messages:
            response_data = {"message": message}
            response = AddGoldenConfigNodeResponse(**response_data)

            assert response.message == message


class TestRenderTemplateResponse:
    """Test cases for RenderTemplateResponse model."""

    def test_create_valid_response(self):
        """Test creating a valid RenderTemplateResponse."""
        response_data = {
            "result": "interface GigabitEthernet0/0\n ip address 192.168.1.1 255.255.255.0"
        }

        response = RenderTemplateResponse(**response_data)

        assert (
            response.result
            == "interface GigabitEthernet0/0\n ip address 192.168.1.1 255.255.255.0"
        )

    def test_create_response_with_various_results(self):
        """Test creating response with various rendered template results."""
        results = [
            "hostname {{ hostname }}",
            "interface Loopback0\n description Management Interface",
            "<configuration><interface name='ge-0/0/0'><unit name='0'></unit></interface></configuration>",
            "router ospf 1\n network 0.0.0.0 255.255.255.255 area 0",
            "",
        ]

        for result in results:
            response_data = {"result": result}
            response = RenderTemplateResponse(**response_data)

            assert response.result == result

    def test_response_missing_result(self):
        """Test RenderTemplateResponse validation with missing result."""
        response_data = {}

        with pytest.raises(ValidationError) as exc_info:
            RenderTemplateResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "missing"
        assert errors[0]["loc"] == ("result",)

    def test_response_invalid_result_type(self):
        """Test RenderTemplateResponse validation with invalid result type."""
        response_data = {"result": 12345}

        with pytest.raises(ValidationError) as exc_info:
            RenderTemplateResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "string_type"
        assert errors[0]["loc"] == ("result",)

    def test_response_none_result(self):
        """Test RenderTemplateResponse validation with None result."""
        response_data = {"result": None}

        with pytest.raises(ValidationError) as exc_info:
            RenderTemplateResponse(**response_data)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "string_type"
        assert errors[0]["loc"] == ("result",)

    def test_response_empty_result(self):
        """Test RenderTemplateResponse with empty result string."""
        response_data = {"result": ""}

        response = RenderTemplateResponse(**response_data)

        assert response.result == ""

    def test_response_serialization(self):
        """Test RenderTemplateResponse serialization."""
        response_data = {"result": "vlan 10\n name Production\n exit"}

        response = RenderTemplateResponse(**response_data)
        serialized = response.model_dump()

        assert serialized == response_data

    def test_response_json_serialization(self):
        """Test RenderTemplateResponse JSON serialization."""
        response_data = {"result": "access-list 100 permit tcp any any eq 80"}

        response = RenderTemplateResponse(**response_data)
        json_str = response.model_dump_json()

        assert '"result":"access-list 100 permit tcp any any eq 80"' in json_str

    def test_response_with_multiline_template(self):
        """Test RenderTemplateResponse with multiline template results."""
        multiline_result = """hostname Router1
!
interface GigabitEthernet0/0
 ip address 10.1.1.1 255.255.255.0
 no shutdown
!
router bgp 65001
 neighbor 10.1.1.2 remote-as 65002
 address-family ipv4
  network 10.1.1.0 mask 255.255.255.0
 exit-address-family"""

        response_data = {"result": multiline_result}
        response = RenderTemplateResponse(**response_data)

        assert response.result == multiline_result
        assert "hostname Router1" in response.result
        assert "router bgp 65001" in response.result

    def test_response_with_special_characters(self):
        """Test RenderTemplateResponse with special characters in result."""
        special_results = [
            "password $1$abcd$efghijklmnop",
            "description 'Main uplink to provider (100% redundant)'",
            "match ip address prefix-list PRIVATE-NETS",
            "crypto isakmp key Th!s1sMyK3y address 192.168.1.1",
        ]

        for result in special_results:
            response_data = {"result": result}
            response = RenderTemplateResponse(**response_data)

            assert response.result == result



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest
from unittest.mock import AsyncMock, MagicMock, patch

from pydantic import BaseModel

from itential_mcp.bindings.endpoint import _get_trigger, start_workflow, new
from itential_mcp.platform import PlatformClient
from itential_mcp.core import exceptions
from fastmcp import Context


class MockResponse(BaseModel):
    """Mock response model for testing."""

    data: dict


@pytest.fixture
def mock_platform_client():
    """Mock PlatformClient for testing."""
    client_mock = AsyncMock(spec=PlatformClient)
    client_mock.client = AsyncMock()
    return client_mock


@pytest.fixture
def mock_endpoint_tool():
    """Mock EndpointTool configuration for testing."""
    tool = MagicMock()
    tool.automation = "test-automation"
    tool.name = "test-trigger"
    return tool


@pytest.fixture
def mock_context():
    """Mock FastMCP Context for testing."""
    context = MagicMock(spec=Context)
    context.request_context.lifespan_context.get.return_value = AsyncMock(
        spec=PlatformClient
    )
    return context


@pytest.fixture
def mock_automation_response():
    """Mock response for automation API call."""
    return {
        "metadata": {"total": 1},
        "data": [{"_id": "automation-123", "name": "test-automation"}],
    }


@pytest.fixture
def mock_triggers_response():
    """Mock response for triggers API call."""
    return {
        "data": [
            {
                "name": "test-trigger",
                "routeName": "test-route",
                "description": "Test trigger",
                "schema": {"type": "object"},
            },
            {
                "name": "other-trigger",
                "routeName": "other-route",
                "description": "Other trigger",
                "schema": {"type": "object"},
            },
        ]
    }


@pytest.fixture
def mock_trigger_data():
    """Mock trigger data for testing."""
    return {
        "name": "test-trigger",
        "routeName": "test-route",
        "description": "Test trigger description",
        "schema": {"type": "object", "properties": {"param": {"type": "string"}}},
    }


class TestGetTrigger:
    """Test cases for _get_trigger function."""

    @pytest.mark.asyncio
    async def test_get_trigger_success(
        self,
        mock_platform_client,
        mock_endpoint_tool,
        mock_automation_response,
        mock_triggers_response,
    ):
        """Test successful trigger retrieval."""
        # Setup
        mock_platform_client.client.get.side_effect = [
            # First call for automation
            MagicMock(json=lambda: mock_automation_response),
            # Second call for triggers
            MagicMock(json=lambda: mock_triggers_response),
        ]

        # Execute
        result = await _get_trigger(mock_platform_client, mock_endpoint_tool)

        # Verify
        assert result == mock_triggers_response["data"][0]
        assert mock_platform_client.client.get.call_count == 2

        # Verify first call (automation lookup)
        mock_platform_client.client.get.assert_any_call(
            "/operations-manager/automations",
            params={"equals": "test-automation", "equalsField": "name"},
        )

        # Verify second call (triggers lookup)
        mock_platform_client.client.get.assert_any_call(
            "/operations-manager/triggers",
            params={"equals": "automation-123", "equalsField": "actionId"},
        )

    @pytest.mark.asyncio
    async def test_get_trigger_automation_not_found(
        self, mock_platform_client, mock_endpoint_tool
    ):
        """Test _get_trigger raises NotFoundError when automation is not found."""
        # Setup - automation not found
        automation_response = {"metadata": {"total": 0}, "data": []}
        mock_platform_client.client.get.return_value = MagicMock(
            json=lambda: automation_response
        )

        # Execute and verify
        with pytest.raises(
            exceptions.NotFoundError,
            match="automation test-automation could not be found",
        ):
            await _get_trigger(mock_platform_client, mock_endpoint_tool)

        # Verify only one call was made (automation lookup)
        assert mock_platform_client.client.get.call_count == 1

    @pytest.mark.asyncio
    async def test_get_trigger_multiple_automations_found(
        self, mock_platform_client, mock_endpoint_tool
    ):
        """Test _get_trigger with multiple automations - uses first match, then fails on trigger lookup."""
        # Setup - multiple automations found, but empty triggers response
        automation_response = {
            "metadata": {"total": 2},
            "data": [
                {"_id": "automation-123", "name": "test-automation"},
                {"_id": "automation-456", "name": "test-automation"},
            ],
        }
        triggers_response = {"data": []}

        # Mock responses in order: automation lookup succeeds, trigger lookup returns empty
        mock_platform_client.client.get.side_effect = [
            MagicMock(json=lambda: automation_response),
            MagicMock(json=lambda: triggers_response),
        ]

        # Execute and verify - should fail when trigger is not found
        with pytest.raises(
            exceptions.NotFoundError,
            match="trigger test-trigger could not be found",
        ):
            await _get_trigger(mock_platform_client, mock_endpoint_tool)

    @pytest.mark.asyncio
    async def test_get_trigger_trigger_not_found(
        self, mock_platform_client, mock_endpoint_tool, mock_automation_response
    ):
        """Test _get_trigger raises NotFoundError when trigger is not found."""
        # Setup
        triggers_response = {
            "data": [{"name": "other-trigger", "routeName": "other-route"}]
        }
        mock_platform_client.client.get.side_effect = [
            MagicMock(json=lambda: mock_automation_response),
            MagicMock(json=lambda: triggers_response),
        ]

        # Execute and verify
        with pytest.raises(
            exceptions.NotFoundError, match="trigger test-trigger could not be found"
        ):
            await _get_trigger(mock_platform_client, mock_endpoint_tool)

    @pytest.mark.asyncio
    async def test_get_trigger_empty_triggers(
        self, mock_platform_client, mock_endpoint_tool, mock_automation_response
    ):
        """Test _get_trigger raises NotFoundError when no triggers are returned."""
        # Setup
        triggers_response = {"data": []}
        mock_platform_client.client.get.side_effect = [
            MagicMock(json=lambda: mock_automation_response),
            MagicMock(json=lambda: triggers_response),
        ]

        # Execute and verify
        with pytest.raises(
            exceptions.NotFoundError, match="trigger test-trigger could not be found"
        ):
            await _get_trigger(mock_platform_client, mock_endpoint_tool)


class TestStartWorkflow:
    """Test cases for start_workflow function."""

    @patch("itential_mcp.bindings.endpoint._get_trigger")
    @patch("itential_mcp.bindings.endpoint.operations_manager.start_workflow")
    @pytest.mark.asyncio
    async def test_start_workflow_success(
        self, mock_ops_start_workflow, mock_get_trigger, mock_context, mock_trigger_data
    ):
        """Test successful workflow start."""
        # Setup
        mock_tool_config = MagicMock()
        mock_data = {"param": "value"}
        mock_get_trigger.return_value = mock_trigger_data
        mock_ops_start_workflow.return_value = MockResponse(data={"job_id": "job-123"})

        # Execute
        result = await start_workflow(mock_context, mock_tool_config, mock_data)

        # Verify
        mock_get_trigger.assert_called_once_with(
            mock_context.request_context.lifespan_context.get.return_value,
            mock_tool_config,
        )
        mock_ops_start_workflow.assert_called_once_with(
            mock_context, route_name="test-route", data=mock_data
        )
        assert result == MockResponse(data={"job_id": "job-123"})

    @patch("itential_mcp.bindings.endpoint._get_trigger")
    @patch("itential_mcp.bindings.endpoint.operations_manager.start_workflow")
    @pytest.mark.asyncio
    async def test_start_workflow_no_data(
        self, mock_ops_start_workflow, mock_get_trigger, mock_context, mock_trigger_data
    ):
        """Test workflow start without input data."""
        # Setup
        mock_tool_config = MagicMock()
        mock_get_trigger.return_value = mock_trigger_data
        mock_ops_start_workflow.return_value = MockResponse(data={"job_id": "job-456"})

        # Execute
        result = await start_workflow(mock_context, mock_tool_config, None)

        # Verify
        mock_ops_start_workflow.assert_called_once_with(
            mock_context, route_name="test-route", data=None
        )
        assert result == MockResponse(data={"job_id": "job-456"})

    @patch("itential_mcp.bindings.endpoint._get_trigger")
    @pytest.mark.asyncio
    async def test_start_workflow_trigger_not_found(
        self, mock_get_trigger, mock_context
    ):
        """Test start_workflow propagates NotFoundError from _get_trigger."""
        # Setup
        mock_tool_config = MagicMock()
        mock_get_trigger.side_effect = exceptions.NotFoundError("trigger not found")

        # Execute and verify
        with pytest.raises(exceptions.NotFoundError, match="trigger not found"):
            await start_workflow(mock_context, mock_tool_config, {})

    @patch("itential_mcp.bindings.endpoint._get_trigger")
    @patch("itential_mcp.bindings.endpoint.operations_manager.start_workflow")
    @pytest.mark.asyncio
    async def test_start_workflow_operations_manager_error(
        self, mock_ops_start_workflow, mock_get_trigger, mock_context, mock_trigger_data
    ):
        """Test start_workflow propagates errors from operations_manager."""
        # Setup
        mock_tool_config = MagicMock()
        mock_get_trigger.return_value = mock_trigger_data
        mock_ops_start_workflow.side_effect = Exception("Workflow execution failed")

        # Execute and verify
        with pytest.raises(Exception, match="Workflow execution failed"):
            await start_workflow(mock_context, mock_tool_config, {})


class TestNew:
    """Test cases for new function."""

    @patch("itential_mcp.bindings.endpoint._get_trigger")
    @pytest.mark.asyncio
    async def test_new_success(
        self,
        mock_get_trigger,
        mock_endpoint_tool,
        mock_platform_client,
        mock_trigger_data,
    ):
        """Test successful creation of bound workflow function."""
        # Setup
        mock_get_trigger.return_value = mock_trigger_data

        # Execute
        fn, description = await new(mock_endpoint_tool, mock_platform_client)

        # Verify
        mock_get_trigger.assert_called_once_with(
            mock_platform_client, mock_endpoint_tool
        )
        assert fn == start_workflow
        assert "Test trigger description" in description
        assert "Args:" in description
        assert "data (dict):" in description

    @patch("itential_mcp.bindings.endpoint._get_trigger")
    @pytest.mark.asyncio
    async def test_new_empty_description(
        self, mock_get_trigger, mock_endpoint_tool, mock_platform_client
    ):
        """Test new function with empty trigger description."""
        # Setup
        trigger_data = {
            "name": "test-trigger",
            "routeName": "test-route",
            "description": "",
            "schema": {"type": "object"},
        }
        mock_get_trigger.return_value = trigger_data

        # Execute
        fn, description = await new(mock_endpoint_tool, mock_platform_client)

        # Verify
        assert fn == start_workflow
        assert "Args:\ndata (dict):" in description

    @patch("itential_mcp.bindings.endpoint._get_trigger")
    @pytest.mark.asyncio
    async def test_new_none_description(
        self, mock_get_trigger, mock_endpoint_tool, mock_platform_client
    ):
        """Test new function with None trigger description."""
        # Setup
        trigger_data = {
            "name": "test-trigger",
            "routeName": "test-route",
            "description": None,
            "schema": {"type": "object"},
        }
        mock_get_trigger.return_value = trigger_data

        # Execute
        fn, description = await new(mock_endpoint_tool, mock_platform_client)

        # Verify
        assert fn == start_workflow
        assert "Args:\ndata (dict):" in description

    @patch("itential_mcp.bindings.endpoint._get_trigger")
    @pytest.mark.asyncio
    async def test_new_trigger_not_found(
        self, mock_get_trigger, mock_endpoint_tool, mock_platform_client
    ):
        """Test new function propagates NotFoundError from _get_trigger."""
        # Setup
        mock_get_trigger.side_effect = exceptions.NotFoundError("trigger not found")

        # Execute and verify
        with pytest.raises(exceptions.NotFoundError, match="trigger not found"):
            await new(mock_endpoint_tool, mock_platform_client)


class TestEndpointIntegration:
    """Integration tests for endpoint module."""

    @pytest.mark.asyncio
    async def test_full_endpoint_workflow(
        self, mock_platform_client, mock_endpoint_tool
    ):
        """Test complete endpoint workflow from trigger retrieval to function binding."""
        # Setup
        automation_response = {
            "metadata": {"total": 1},
            "data": [{"_id": "automation-123", "name": "test-automation"}],
        }
        triggers_response = {
            "data": [
                {
                    "name": "test-trigger",
                    "routeName": "test-workflow-route",
                    "description": "Complete test workflow",
                    "schema": {
                        "type": "object",
                        "properties": {"device": {"type": "string"}},
                    },
                }
            ]
        }

        # We need to provide enough responses for both calls to _get_trigger
        # (once in new() and once directly in test)
        mock_platform_client.client.get.side_effect = [
            MagicMock(json=lambda: automation_response),
            MagicMock(json=lambda: triggers_response),
            MagicMock(json=lambda: automation_response),
            MagicMock(json=lambda: triggers_response),
        ]

        # Execute
        fn, description = await new(mock_endpoint_tool, mock_platform_client)

        # Verify
        assert fn == start_workflow
        assert "Complete test workflow" in description
        assert "Args:" in description
        assert "data (dict):" in description
        assert "'type': 'object'" in description

        # Verify the trigger was retrieved correctly in a separate call
        trigger = await _get_trigger(mock_platform_client, mock_endpoint_tool)
        assert trigger["name"] == "test-trigger"
        assert trigger["routeName"] == "test-workflow-route"
        assert trigger["description"] == "Complete test workflow"

    @pytest.mark.asyncio
    async def test_endpoint_error_handling_chain(
        self, mock_platform_client, mock_endpoint_tool
    ):
        """Test that errors propagate correctly through the endpoint chain."""
        # Setup - simulate automation not found
        automation_response = {"metadata": {"total": 0}, "data": []}
        mock_platform_client.client.get.return_value = MagicMock(
            json=lambda: automation_response
        )

        # Test error propagation through _get_trigger
        with pytest.raises(exceptions.NotFoundError):
            await _get_trigger(mock_platform_client, mock_endpoint_tool)

        # Test error propagation through new function
        with pytest.raises(exceptions.NotFoundError):
            await new(mock_endpoint_tool, mock_platform_client)



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from starlette.requests import Request
from starlette.responses import JSONResponse

from ..platform import PlatformClient


async def get_healthz(request: Request) -> JSONResponse:
    """
    Kubernetes healthz endpoint - general health check.

    Returns HTTP 200 if the server is healthy, otherwise HTTP 503.
    This is used by Kubernetes for general health monitoring.

    Args:
        request: The incoming HTTP request

    Returns:
        JSONResponse: Status response with HTTP 200 for healthy, 503 for unhealthy
    """
    try:
        # Basic server health check - server is running if we reach here
        return JSONResponse(content={"status": "ok"}, status_code=200)
    except Exception:
        return JSONResponse(content={"status": "unhealthy"}, status_code=503)


async def get_readyz(request: Request) -> JSONResponse:
    """
    Kubernetes readyz endpoint - readiness probe.

    Returns HTTP 200 if the server is ready to accept traffic, otherwise HTTP 503.
    Used by Kubernetes to determine if the pod should receive traffic.

    This endpoint checks if the server can successfully communicate with the
    Itential Platform by performing a /whoami request. If successful, the
    server is considered ready to handle requests.

    Args:
        request: The incoming HTTP request

    Returns:
        JSONResponse: Status response with HTTP 200 for ready, 503 for not ready
    """
    try:
        async with PlatformClient() as client_instance:
            await client_instance.get("/whoami")
        return JSONResponse(content={"status": "ready"}, status_code=200)
    except Exception as e:
        return JSONResponse(
            content={"status": "not ready", "reason": str(e)}, status_code=503
        )


async def get_livez(request: Request) -> JSONResponse:
    """
    Kubernetes livez endpoint - liveness probe.

    Returns HTTP 200 if the server is alive, otherwise HTTP 503.
    Used by Kubernetes to determine if the pod should be restarted.

    Args:
        request: The incoming HTTP request

    Returns:
        JSONResponse: Status response with HTTP 200 for alive, 503 for dead
    """
    try:
        # Basic liveness check - if we can respond, we're alive
        return JSONResponse(content={"status": "alive"}, status_code=200)
    except Exception:
        return JSONResponse(content={"status": "dead"}, status_code=503)



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest
from unittest.mock import AsyncMock, MagicMock, patch

from itential_mcp.core import exceptions
from itential_mcp.platform.services.applications import Service


class TestApplicationsService:
    """Test cases for the applications Service class"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    def test_service_name(self, mock_client):
        """Test that the service has the correct name"""
        service = Service(mock_client)
        assert service.name == "applications"

    @pytest.mark.asyncio
    async def test_get_application_health_success(self, service, mock_client):
        """Test successful application health retrieval"""
        # Mock response data
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "total": 1,
            "results": [
                {
                    "results": [
                        {
                            "id": "test-application",
                            "state": "RUNNING",
                            "version": "1.0.0",
                        }
                    ]
                }
            ],
        }
        mock_client.get.return_value = mock_response

        result = await service._get_application_health("test-application")

        # Verify client was called with correct parameters
        mock_client.get.assert_called_once_with(
            "/health/applications",
            params={"equals": "test-application", "equalsField": "id"},
        )

        # Verify result - should return data["results"][0]
        assert result["results"][0]["id"] == "test-application"
        assert result["results"][0]["state"] == "RUNNING"

    @pytest.mark.asyncio
    async def test_get_application_health_not_found(self, service, mock_client):
        """Test application not found scenario"""
        # Mock response with no results
        mock_response = MagicMock()
        mock_response.json.return_value = {"total": 0, "results": []}
        mock_client.get.return_value = mock_response

        with pytest.raises(exceptions.NotFoundError) as exc_info:
            await service._get_application_health("nonexistent-application")

        assert "unable to find application nonexistent-application" in str(
            exc_info.value
        )

    @pytest.mark.asyncio
    async def test_get_application_health_multiple_results(self, service, mock_client):
        """Test scenario where multiple applications are found (should not happen)"""
        # Mock response with multiple results
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "total": 2,
            "results": [
                {"results": [{"id": "app1", "state": "RUNNING"}]},
                {"results": [{"id": "app2", "state": "STOPPED"}]},
            ],
        }
        mock_client.get.return_value = mock_response

        with pytest.raises(exceptions.NotFoundError):
            await service._get_application_health("duplicate-application")


class TestStartApplication:
    """Test cases for the start_application method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_start_application_already_running(self, service):
        """Test starting an application that's already running"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "RUNNING"}]}

            result = await service.start_application("test-application", 10)

            # Should return immediately without calling PUT
            service.client.put.assert_not_called()
            # Should return the health data
            assert result["results"][0]["state"] == "RUNNING"

    @pytest.mark.asyncio
    async def test_start_application_from_stopped_success(self, service):
        """Test successfully starting a stopped application"""
        with patch.object(service, "_get_application_health") as mock_health:
            # First call returns STOPPED, second returns RUNNING
            mock_health.side_effect = [
                {"results": [{"state": "STOPPED"}]},
                {"results": [{"state": "RUNNING"}]},
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock):
                result = await service.start_application("test-application", 10)

            # Should call PUT to start application
            service.client.put.assert_called_once_with(
                "/applications/test-application/start"
            )

            # Should check health twice (initial + after start)
            assert mock_health.call_count == 2

            # Should return the final health data
            assert result["results"][0]["state"] == "RUNNING"

    @pytest.mark.asyncio
    async def test_start_application_timeout(self, service):
        """Test application start timeout scenario"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Always return STOPPED (never transitions to RUNNING)
            mock_health.side_effect = [
                {"results": [{"state": "STOPPED"}]},  # Initial state
                {"results": [{"state": "STOPPED"}]},  # After start attempt
                {"results": [{"state": "STOPPED"}]},  # Still stopped...
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
                with pytest.raises(exceptions.TimeoutExceededError):
                    await service.start_application("test-application", 2)

                # Should have slept timeout number of times
                assert mock_sleep.call_count == 2

    @pytest.mark.asyncio
    async def test_start_application_dead_state(self, service):
        """Test starting an application in DEAD state"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "DEAD"}]}

            with pytest.raises(exceptions.InvalidStateError) as exc_info:
                await service.start_application("test-application", 10)

            assert "application `test-application` is `DEAD`" in str(exc_info.value)
            service.client.put.assert_not_called()

    @pytest.mark.asyncio
    async def test_start_application_deleted_state(self, service):
        """Test starting an application in DELETED state"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "DELETED"}]}

            with pytest.raises(exceptions.InvalidStateError) as exc_info:
                await service.start_application("test-application", 10)

            assert "application `test-application` is `DELETED`" in str(exc_info.value)


class TestStopApplication:
    """Test cases for the stop_application method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_stop_application_already_stopped(self, service):
        """Test stopping an application that's already stopped"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "STOPPED"}]}

            result = await service.stop_application("test-application", 10)

            # Should return immediately without calling PUT
            service.client.put.assert_not_called()
            # Should return the health data
            assert result["results"][0]["state"] == "STOPPED"

    @pytest.mark.asyncio
    async def test_stop_application_from_running_success(self, service):
        """Test successfully stopping a running application"""
        with patch.object(service, "_get_application_health") as mock_health:
            # First call returns RUNNING, second returns STOPPED
            mock_health.side_effect = [
                {"results": [{"state": "RUNNING"}]},
                {"results": [{"state": "STOPPED"}]},
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock):
                result = await service.stop_application("test-application", 10)

            # Should call PUT to stop application
            service.client.put.assert_called_once_with(
                "/applications/test-application/stop"
            )

            # Should return the final health data
            assert result["results"][0]["state"] == "STOPPED"

    @pytest.mark.asyncio
    async def test_stop_application_timeout(self, service):
        """Test application stop timeout scenario"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Always return RUNNING (never transitions to STOPPED)
            mock_health.side_effect = [
                {"results": [{"state": "RUNNING"}]},  # Initial state
                {"results": [{"state": "RUNNING"}]},  # After stop attempt
                {"results": [{"state": "RUNNING"}]},  # Still running...
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
                with pytest.raises(exceptions.TimeoutExceededError):
                    await service.stop_application("test-application", 2)

                # Should have slept timeout number of times
                assert mock_sleep.call_count == 2

    @pytest.mark.asyncio
    async def test_stop_application_dead_state(self, service):
        """Test stopping an application in DEAD state"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "DEAD"}]}

            with pytest.raises(exceptions.InvalidStateError) as exc_info:
                await service.stop_application("test-application", 10)

            assert "application `test-application` is `DEAD`" in str(exc_info.value)
            service.client.put.assert_not_called()

    @pytest.mark.asyncio
    async def test_stop_application_deleted_state(self, service):
        """Test stopping an application in DELETED state"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "DELETED"}]}

            with pytest.raises(exceptions.InvalidStateError) as exc_info:
                await service.stop_application("test-application", 10)

            assert "application `test-application` is `DELETED`" in str(exc_info.value)
            service.client.put.assert_not_called()

    @pytest.mark.asyncio
    async def test_stop_application_multiple_pages(self, service):
        """Test stopping application with pagination during wait loop"""
        with patch.object(service, "_get_application_health") as mock_health:
            # First call returns RUNNING, then STOPPED after several checks
            mock_health.side_effect = [
                {"results": [{"state": "RUNNING"}]},  # Initial check
                {"results": [{"state": "RUNNING"}]},  # First wait check
                {"results": [{"state": "RUNNING"}]},  # Second wait check
                {"results": [{"state": "STOPPED"}]},  # Finally stopped
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock):
                result = await service.stop_application("test-application", 10)

            # Should call PUT to stop application
            service.client.put.assert_called_once_with(
                "/applications/test-application/stop"
            )

            # Should check health multiple times
            assert mock_health.call_count == 4
            assert result["results"][0]["state"] == "STOPPED"


class TestRestartApplication:
    """Test cases for the restart_application method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_restart_application_from_running_success(self, service):
        """Test successfully restarting a running application"""
        with patch.object(service, "_get_application_health") as mock_health:
            # First call returns RUNNING, second returns RUNNING after restart
            mock_health.side_effect = [
                {"results": [{"state": "RUNNING"}]},
                {"results": [{"state": "RUNNING"}]},
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock):
                result = await service.restart_application("test-application", 10)

            # Should call PUT to restart application
            service.client.put.assert_called_once_with(
                "/applications/test-application/restart"
            )

            # Should return the final health data
            assert result["results"][0]["state"] == "RUNNING"

    @pytest.mark.asyncio
    async def test_restart_application_stopped_state(self, service):
        """Test restarting an application in STOPPED state"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "STOPPED"}]}

            with pytest.raises(exceptions.InvalidStateError) as exc_info:
                await service.restart_application("test-application", 10)

            assert "application `test-application` is `STOPPED`" in str(exc_info.value)
            service.client.put.assert_not_called()

    @pytest.mark.asyncio
    async def test_restart_application_dead_state(self, service):
        """Test restarting an application in DEAD state"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "DEAD"}]}

            with pytest.raises(exceptions.InvalidStateError) as exc_info:
                await service.restart_application("test-application", 10)

            assert "application `test-application` is `DEAD`" in str(exc_info.value)
            service.client.put.assert_not_called()

    @pytest.mark.asyncio
    async def test_restart_application_deleted_state(self, service):
        """Test restarting an application in DELETED state"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "DELETED"}]}

            with pytest.raises(exceptions.InvalidStateError) as exc_info:
                await service.restart_application("test-application", 10)

            assert "application `test-application` is `DELETED`" in str(exc_info.value)
            service.client.put.assert_not_called()

    @pytest.mark.asyncio
    async def test_restart_application_timeout(self, service):
        """Test application restart timeout scenario"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Always return RUNNING at first, then stays in intermediate state
            mock_health.side_effect = [
                {"results": [{"state": "RUNNING"}]},  # Initial state
                {"results": [{"state": "RESTARTING"}]},  # After restart attempt
                {"results": [{"state": "RESTARTING"}]},  # Still restarting...
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
                with pytest.raises(exceptions.TimeoutExceededError):
                    await service.restart_application("test-application", 2)

                # Should have slept timeout number of times
                assert mock_sleep.call_count == 2

    @pytest.mark.asyncio
    async def test_restart_application_multiple_state_changes(self, service):
        """Test restarting application through multiple state changes"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Simulate application going through restart states
            mock_health.side_effect = [
                {"results": [{"state": "RUNNING"}]},  # Initial check
                {"results": [{"state": "RESTARTING"}]},  # First wait check
                {"results": [{"state": "STARTING"}]},  # Second wait check
                {"results": [{"state": "RUNNING"}]},  # Finally running
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock):
                result = await service.restart_application("test-application", 10)

            # Should call PUT to restart application
            service.client.put.assert_called_once_with(
                "/applications/test-application/restart"
            )

            # Should check health multiple times
            assert mock_health.call_count == 4
            assert result["results"][0]["state"] == "RUNNING"

    @pytest.mark.asyncio
    async def test_restart_application_already_running_no_change_needed(self, service):
        """Test restarting application that immediately returns to RUNNING"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Application is RUNNING before and after restart
            mock_health.side_effect = [
                {"results": [{"state": "RUNNING"}]},  # Initial check
                {
                    "results": [{"state": "RUNNING"}]
                },  # Immediately running after restart
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock):
                result = await service.restart_application("test-application", 10)

            # Should still call PUT to restart
            service.client.put.assert_called_once_with(
                "/applications/test-application/restart"
            )

            # Should return running state
            assert result["results"][0]["state"] == "RUNNING"


class TestServiceIntegration:
    """Integration tests for the Service class"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_service_inherits_from_servicebase(self, service):
        """Test that Service properly inherits from ServiceBase"""
        from itential_mcp.platform.services import ServiceBase

        assert isinstance(service, ServiceBase)

    @pytest.mark.asyncio
    async def test_method_signatures_consistency(self, service):
        """Test that all methods have consistent parameter signatures"""
        import inspect

        # Check start_application signature
        start_sig = inspect.signature(service.start_application)
        assert "name" in start_sig.parameters
        assert "timeout" in start_sig.parameters

        # Check stop_application signature
        stop_sig = inspect.signature(service.stop_application)
        assert "name" in stop_sig.parameters
        assert "timeout" in stop_sig.parameters

        # Check restart_application signature
        restart_sig = inspect.signature(service.restart_application)
        assert "name" in restart_sig.parameters
        assert "timeout" in restart_sig.parameters


class TestErrorHandling:
    """Test error handling scenarios"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_client_connection_error(self, service, mock_client):
        """Test handling of client connection errors"""
        mock_client.get.side_effect = Exception("Connection failed")

        with pytest.raises(Exception, match="Connection failed"):
            await service._get_application_health("test-application")

    @pytest.mark.asyncio
    async def test_malformed_response_handling(self, service, mock_client):
        """Test handling of malformed API responses"""
        # Mock response with missing 'total' field
        mock_response = MagicMock()
        mock_response.json.return_value = {"results": []}
        mock_client.get.return_value = mock_response

        with pytest.raises(KeyError):
            await service._get_application_health("test-application")


class TestEdgeCases:
    """Test edge cases and boundary conditions"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_application_name_with_special_characters(self, service, mock_client):
        """Test application names with special characters"""
        special_name = "test-application_123.456@domain"

        mock_response = MagicMock()
        mock_response.json.return_value = {
            "total": 1,
            "results": [{"results": [{"state": "RUNNING"}]}],
        }
        mock_client.get.return_value = mock_response

        await service._get_application_health(special_name)

        # Verify the special name was used correctly in the API call
        mock_client.get.assert_called_with(
            "/health/applications", params={"equals": special_name, "equalsField": "id"}
        )

    @pytest.mark.asyncio
    async def test_unicode_application_name(self, service, mock_client):
        """Test application names with Unicode characters"""
        unicode_name = "测试应用程序"

        mock_response = MagicMock()
        mock_response.json.return_value = {
            "total": 1,
            "results": [{"results": [{"state": "RUNNING"}]}],
        }
        mock_client.get.return_value = mock_response

        await service._get_application_health(unicode_name)

        # Verify Unicode name was handled correctly
        mock_client.get.assert_called_with(
            "/health/applications", params={"equals": unicode_name, "equalsField": "id"}
        )

    @pytest.mark.asyncio
    async def test_zero_timeout_behavior(self, service, mock_client):
        """Test behavior with zero timeout"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.return_value = {"results": [{"state": "STOPPED"}]}

            with pytest.raises(exceptions.TimeoutExceededError):
                await service.start_application("test-application", 0)

        # Should call PUT but timeout immediately since timeout=0
        service.client.put.assert_called_once_with(
            "/applications/test-application/start"
        )

    @pytest.mark.asyncio
    async def test_negative_timeout_behavior(self, service, mock_client):
        """Test behavior with negative timeout (demonstrates infinite loop issue)"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Return STOPPED first, then RUNNING after first attempt
            mock_health.side_effect = [
                {"results": [{"state": "STOPPED"}]},
                {
                    "results": [{"state": "RUNNING"}]
                },  # Immediately running to break loop
            ]

            # Should complete successfully with negative timeout if app starts immediately
            result = await service.start_application("test-application", -1)

        # Should call PUT and succeed if application transitions to RUNNING immediately
        service.client.put.assert_called_once_with(
            "/applications/test-application/start"
        )
        assert result["results"][0]["state"] == "RUNNING"

    @pytest.mark.asyncio
    async def test_large_timeout_value(self, service, mock_client):
        """Test behavior with very large timeout value"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Return STOPPED first, then RUNNING
            mock_health.side_effect = [
                {"results": [{"state": "STOPPED"}]},
                {"results": [{"state": "RUNNING"}]},
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock):
                result = await service.start_application("test-application", 999999)

            # Should work normally with large timeout
            assert result["results"][0]["state"] == "RUNNING"

    @pytest.mark.asyncio
    async def test_start_application_not_found_error_propagation(self, service):
        """Test that NotFoundError from _get_application_health is properly propagated"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.side_effect = exceptions.NotFoundError("Application not found")

            with pytest.raises(exceptions.NotFoundError, match="Application not found"):
                await service.start_application("nonexistent-app", 10)

    @pytest.mark.asyncio
    async def test_stop_application_not_found_error_propagation(self, service):
        """Test that NotFoundError from _get_application_health is properly propagated"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.side_effect = exceptions.NotFoundError("Application not found")

            with pytest.raises(exceptions.NotFoundError, match="Application not found"):
                await service.stop_application("nonexistent-app", 10)

    @pytest.mark.asyncio
    async def test_restart_application_not_found_error_propagation(self, service):
        """Test that NotFoundError from _get_application_health is properly propagated"""
        with patch.object(service, "_get_application_health") as mock_health:
            mock_health.side_effect = exceptions.NotFoundError("Application not found")

            with pytest.raises(exceptions.NotFoundError, match="Application not found"):
                await service.restart_application("nonexistent-app", 10)

    @pytest.mark.asyncio
    async def test_concurrent_application_operations(self, service):
        """Test that service methods can handle concurrent operations"""
        import asyncio

        with patch.object(service, "_get_application_health") as mock_health:
            # Mock different responses for different calls
            mock_health.side_effect = [
                {"results": [{"state": "RUNNING"}]},  # For first operation
                {"results": [{"state": "STOPPED"}]},  # For second operation
            ]

            # Run two operations concurrently (though they should be independent)
            task1 = service.start_application("app1", 10)
            task2 = service.stop_application("app2", 10)

            # Both should complete without interfering with each other
            result1, result2 = await asyncio.gather(task1, task2)

            # Verify results
            assert result1["results"][0]["state"] == "RUNNING"
            assert result2["results"][0]["state"] == "STOPPED"


class TestApplicationStateTransitions:
    """Test comprehensive application state transition scenarios"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_complex_restart_state_transition(self, service):
        """Test restart with complex state transitions"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Simulate realistic restart state progression
            mock_health.side_effect = [
                {"results": [{"state": "RUNNING"}]},  # Initial state
                {"results": [{"state": "STOPPING"}]},  # Stopping phase
                {"results": [{"state": "STOPPED"}]},  # Stopped phase
                {"results": [{"state": "STARTING"}]},  # Starting phase
                {"results": [{"state": "RUNNING"}]},  # Finally running
            ]

            with patch("asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
                result = await service.restart_application("test-app", 10)

            # Should go through multiple state checks
            assert mock_health.call_count == 5
            assert (
                mock_sleep.call_count == 3
            )  # Sleep after non-RUNNING states (4 checks - 1 final = 3 sleeps)
            assert result["results"][0]["state"] == "RUNNING"

    @pytest.mark.asyncio
    async def test_start_from_error_state(self, service):
        """Test starting application from an ERROR state"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Application in ERROR state initially - will not trigger PUT call
            mock_health.return_value = {"results": [{"state": "ERROR"}]}

            result = await service.start_application("test-app", 10)

            # Should not attempt to start from ERROR state (only STOPPED triggers PUT)
            service.client.put.assert_not_called()
            assert result["results"][0]["state"] == "ERROR"

    @pytest.mark.asyncio
    async def test_unknown_state_handling(self, service):
        """Test handling of unknown application states"""
        with patch.object(service, "_get_application_health") as mock_health:
            # Unknown state initially - will not trigger PUT call
            mock_health.return_value = {"results": [{"state": "UNKNOWN"}]}

            result = await service.start_application("test-app", 10)

            # Should not attempt to start from UNKNOWN state (only STOPPED triggers PUT)
            service.client.put.assert_not_called()
            assert result["results"][0]["state"] == "UNKNOWN"



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

"""Tests for server keepalive functionality."""

import asyncio
import pytest
from unittest.mock import AsyncMock, Mock

from itential_mcp.server.keepalive import keepalive_task, start_keepalive


class MockPlatformClient:
    def __init__(self):
        self.get = AsyncMock()


class TestKeepaliveTask:
    """Test cases for keepalive_task function."""

    @pytest.mark.asyncio
    async def test_keepalive_task_makes_request(self):
        """Test that keepalive task makes GET requests to /whoami."""
        client = MockPlatformClient()
        mock_response = Mock()
        mock_response.status = 200
        client.get.return_value = mock_response

        # Create task and let it run for a short time
        task = asyncio.create_task(keepalive_task(client, 0.1))
        await asyncio.sleep(0.15)  # Let it run for one iteration
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

        # Verify the client.get was called with /whoami
        client.get.assert_called_with("/whoami")
        assert client.get.call_count >= 1

    @pytest.mark.asyncio
    async def test_keepalive_task_continues_on_error(self):
        """Test that keepalive task continues running even when requests fail."""
        client = MockPlatformClient()
        client.get.side_effect = Exception("Connection failed")

        # Create task and let it run for a short time
        task = asyncio.create_task(keepalive_task(client, 0.1))
        await asyncio.sleep(0.25)  # Let it run for multiple iterations
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

        # Verify multiple calls were made despite errors
        assert client.get.call_count >= 2

    @pytest.mark.asyncio
    async def test_keepalive_task_respects_interval(self):
        """Test that keepalive task waits for the specified interval."""
        client = MockPlatformClient()
        mock_response = Mock()
        mock_response.status = 200
        client.get.return_value = mock_response

        start_time = asyncio.get_event_loop().time()

        # Create task with 0.2 second interval
        task = asyncio.create_task(keepalive_task(client, 0.2))
        await asyncio.sleep(0.45)  # Let it run for 2+ intervals
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

        elapsed_time = asyncio.get_event_loop().time() - start_time

        # Should have made at least 2 calls in 0.45 seconds with 0.2 interval
        assert client.get.call_count >= 2
        assert elapsed_time >= 0.4  # At least 2 intervals


class TestStartKeepalive:
    """Test cases for start_keepalive function."""

    @pytest.mark.asyncio
    async def test_start_keepalive_returns_task(self):
        """Test that start_keepalive returns an asyncio Task."""
        client = MockPlatformClient()

        task = start_keepalive(client, 300)

        assert isinstance(task, asyncio.Task)
        assert not task.done()

        # Clean up the task
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass

    @pytest.mark.asyncio
    async def test_start_keepalive_task_runs(self):
        """Test that the task returned by start_keepalive actually runs."""
        client = MockPlatformClient()
        mock_response = Mock()
        mock_response.status = 200
        client.get.return_value = mock_response

        task = start_keepalive(client, 0.1)

        # Let the task run briefly
        await asyncio.sleep(0.15)
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

        # Verify the keepalive task made at least one request
        client.get.assert_called_with("/whoami")


class TestKeepaliveIntegration:
    """Integration tests for keepalive functionality."""

    @pytest.mark.asyncio
    async def test_keepalive_with_real_async_patterns(self):
        """Test keepalive works with realistic async patterns."""
        client = MockPlatformClient()
        responses = [
            Mock(status=200),
            Mock(status=200),
            Exception("Network error"),
            Mock(status=200),
        ]
        client.get.side_effect = responses

        task = start_keepalive(client, 0.05)  # Very short interval for testing

        # Let it run through multiple iterations including an error
        await asyncio.sleep(0.25)
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

        # Should have attempted multiple calls
        assert client.get.call_count >= 3



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest
from unittest.mock import AsyncMock, MagicMock

from itential_mcp.platform.services.gateway_manager import Service


class TestGatewayManagerService:
    """Test cases for the gateway_manager Service class"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    def test_service_name(self, mock_client):
        """Test that the service has the correct name"""
        service = Service(mock_client)
        assert service.name == "gateway_manager"


class TestGetServices:
    """Test cases for the get_services method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_get_services_success(self, service, mock_client):
        """Test successful services retrieval"""
        expected_services = [
            {
                "name": "test-service-1",
                "cluster": "cluster-1",
                "type": "ansible-playbook",
                "description": "Test Ansible playbook service",
                "decorator": {
                    "type": "object",
                    "properties": {"param1": {"type": "string"}},
                },
            },
            {
                "name": "test-service-2",
                "cluster": "cluster-2",
                "type": "python-script",
                "description": "Test Python script service",
                "decorator": {
                    "type": "object",
                    "properties": {"param2": {"type": "integer"}},
                },
            },
        ]

        # Mock response data
        mock_response = MagicMock()
        mock_response.json.return_value = {"result": expected_services}
        mock_client.get.return_value = mock_response

        result = await service.get_services()

        # Verify client was called with correct endpoint
        mock_client.get.assert_called_once_with("/gateway_manager/v1/services")

        # Verify result
        assert result == expected_services
        assert len(result) == 2
        assert result[0]["name"] == "test-service-1"
        assert result[1]["type"] == "python-script"

    @pytest.mark.asyncio
    async def test_get_services_empty_list(self, service, mock_client):
        """Test services retrieval with empty list"""
        # Mock response with empty list
        mock_response = MagicMock()
        mock_response.json.return_value = {"result": []}
        mock_client.get.return_value = mock_response

        result = await service.get_services()

        assert result == []
        mock_client.get.assert_called_once_with("/gateway_manager/v1/services")

    @pytest.mark.asyncio
    async def test_get_services_client_error(self, service, mock_client):
        """Test services retrieval with client error"""
        mock_client.get.side_effect = Exception("API Error")

        with pytest.raises(Exception, match="API Error"):
            await service.get_services()

    @pytest.mark.asyncio
    async def test_get_services_json_error(self, service, mock_client):
        """Test services retrieval with JSON decode error"""
        mock_response = MagicMock()
        mock_response.json.side_effect = ValueError("Invalid JSON")
        mock_client.get.return_value = mock_response

        with pytest.raises(ValueError, match="Invalid JSON"):
            await service.get_services()


class TestGetGateways:
    """Test cases for the get_gateways method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_get_gateways_success(self, service, mock_client):
        """Test successful gateways retrieval"""
        expected_gateways = [
            {
                "name": "gateway-1",
                "cluster": "cluster-1",
                "description": "Primary gateway",
                "status": "connected",
                "enabled": True,
            },
            {
                "name": "gateway-2",
                "cluster": "cluster-2",
                "description": "Secondary gateway",
                "status": "disconnected",
                "enabled": False,
            },
        ]

        # Mock response data
        mock_response = MagicMock()
        mock_response.json.return_value = expected_gateways
        mock_client.get.return_value = mock_response

        result = await service.get_gateways()

        # Verify client was called with correct endpoint
        mock_client.get.assert_called_once_with("/gateway_manager/v1/gateways")

        # Verify result
        assert result == expected_gateways
        assert len(result) == 2
        assert result[0]["name"] == "gateway-1"
        assert result[0]["enabled"] is True
        assert result[1]["status"] == "disconnected"

    @pytest.mark.asyncio
    async def test_get_gateways_empty_list(self, service, mock_client):
        """Test gateways retrieval with empty list"""
        # Mock response with empty list
        mock_response = MagicMock()
        mock_response.json.return_value = []
        mock_client.get.return_value = mock_response

        result = await service.get_gateways()

        assert result == []
        mock_client.get.assert_called_once_with("/gateway_manager/v1/gateways")

    @pytest.mark.asyncio
    async def test_get_gateways_client_error(self, service, mock_client):
        """Test gateways retrieval with client error"""
        mock_client.get.side_effect = Exception("Connection failed")

        with pytest.raises(Exception, match="Connection failed"):
            await service.get_gateways()

    @pytest.mark.asyncio
    async def test_get_gateways_json_error(self, service, mock_client):
        """Test gateways retrieval with JSON decode error"""
        mock_response = MagicMock()
        mock_response.json.side_effect = ValueError("Invalid JSON")
        mock_client.get.return_value = mock_response

        with pytest.raises(ValueError, match="Invalid JSON"):
            await service.get_gateways()


class TestRunService:
    """Test cases for the run_service method"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_run_service_success_without_params(self, service, mock_client):
        """Test successful service run without input parameters"""
        service_name = "test-service"
        cluster_name = "test-cluster"
        expected_result = {
            "stdout": "Service executed successfully",
            "stderr": "",
            "return_code": 0,
            "start_time": "2025-01-01T10:00:00Z",
            "end_time": "2025-01-01T10:05:00Z",
            "elapsed_time": 300,
        }

        # Mock response data
        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        result = await service.run_service(service_name, cluster_name)

        # Verify client was called with correct parameters
        expected_body = {"serviceName": service_name, "clusterId": cluster_name}
        mock_client.post.assert_called_once_with(
            "/gateway_manager/v1/services/run", json=expected_body
        )

        # Verify result
        assert result == expected_result
        assert result["return_code"] == 0

    @pytest.mark.asyncio
    async def test_run_service_success_with_params(self, service, mock_client):
        """Test successful service run with input parameters"""
        service_name = "test-service"
        cluster_name = "test-cluster"
        input_params = {"param1": "value1", "param2": 42}
        expected_result = {
            "stdout": "Service executed with params",
            "stderr": "",
            "return_code": 0,
            "start_time": "2025-01-01T10:00:00Z",
            "end_time": "2025-01-01T10:05:00Z",
            "elapsed_time": 300,
        }

        # Mock response data
        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        result = await service.run_service(service_name, cluster_name, input_params)

        # Verify client was called with correct parameters including input params
        expected_body = {
            "serviceName": service_name,
            "clusterId": cluster_name,
            "params": input_params,
        }
        mock_client.post.assert_called_once_with(
            "/gateway_manager/v1/services/run", json=expected_body
        )

        # Verify result
        assert result == expected_result

    @pytest.mark.asyncio
    async def test_run_service_with_none_params(self, service, mock_client):
        """Test service run with explicitly None input parameters"""
        service_name = "test-service"
        cluster_name = "test-cluster"
        expected_result = {"return_code": 0}

        # Mock response data
        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        await service.run_service(service_name, cluster_name, None)

        # Verify client was called without params in body
        expected_body = {"serviceName": service_name, "clusterId": cluster_name}
        mock_client.post.assert_called_once_with(
            "/gateway_manager/v1/services/run", json=expected_body
        )

    @pytest.mark.asyncio
    async def test_run_service_with_empty_dict_params(self, service, mock_client):
        """Test service run with empty dict input parameters"""
        service_name = "test-service"
        cluster_name = "test-cluster"
        input_params = {}
        expected_result = {"return_code": 0}

        # Mock response data
        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        await service.run_service(service_name, cluster_name, input_params)

        # Verify client was called without params because empty dict is falsy
        expected_body = {"serviceName": service_name, "clusterId": cluster_name}
        mock_client.post.assert_called_once_with(
            "/gateway_manager/v1/services/run", json=expected_body
        )

    @pytest.mark.asyncio
    async def test_run_service_failure(self, service, mock_client):
        """Test service run failure with non-zero return code"""
        service_name = "failing-service"
        cluster_name = "test-cluster"
        expected_result = {
            "stdout": "",
            "stderr": "Service failed with error",
            "return_code": 1,
            "start_time": "2025-01-01T10:00:00Z",
            "end_time": "2025-01-01T10:01:00Z",
            "elapsed_time": 60,
        }

        # Mock response data
        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        result = await service.run_service(service_name, cluster_name)

        # Should still return the result even if service failed
        assert result == expected_result
        assert result["return_code"] == 1
        assert "error" in result["stderr"]

    @pytest.mark.asyncio
    async def test_run_service_client_error(self, service, mock_client):
        """Test service run with client error"""
        mock_client.post.side_effect = Exception("API Error")

        with pytest.raises(Exception, match="API Error"):
            await service.run_service("test-service", "test-cluster")

    @pytest.mark.asyncio
    async def test_run_service_json_error(self, service, mock_client):
        """Test service run with JSON decode error"""
        mock_response = MagicMock()
        mock_response.json.side_effect = ValueError("Invalid JSON")
        mock_client.post.return_value = mock_response

        with pytest.raises(ValueError, match="Invalid JSON"):
            await service.run_service("test-service", "test-cluster")

    @pytest.mark.asyncio
    async def test_run_service_with_complex_params(self, service, mock_client):
        """Test service run with complex nested input parameters"""
        service_name = "complex-service"
        cluster_name = "test-cluster"
        input_params = {
            "config": {
                "environment": "production",
                "settings": {"timeout": 300, "retries": 3},
            },
            "targets": ["device1", "device2"],
            "variables": {"var1": "value1", "var2": None, "var3": True},
        }
        expected_result = {"return_code": 0}

        # Mock response data
        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        await service.run_service(service_name, cluster_name, input_params)

        # Verify client was called with complex params
        expected_body = {
            "serviceName": service_name,
            "clusterId": cluster_name,
            "params": input_params,
        }
        mock_client.post.assert_called_once_with(
            "/gateway_manager/v1/services/run", json=expected_body
        )


class TestServiceIntegration:
    """Integration tests for the Service class"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_service_inherits_from_servicebase(self, service):
        """Test that Service properly inherits from ServiceBase"""
        from itential_mcp.platform.services import ServiceBase

        assert isinstance(service, ServiceBase)

    @pytest.mark.asyncio
    async def test_method_signatures_consistency(self, service):
        """Test that all methods have consistent parameter signatures"""
        import inspect

        # Check get_services signature
        get_services_sig = inspect.signature(service.get_services)
        assert len(get_services_sig.parameters) == 0

        # Check get_gateways signature
        get_gateways_sig = inspect.signature(service.get_gateways)
        assert len(get_gateways_sig.parameters) == 0

        # Check run_service signature
        run_service_sig = inspect.signature(service.run_service)
        assert "name" in run_service_sig.parameters
        assert "cluster" in run_service_sig.parameters
        assert "input_params" in run_service_sig.parameters

    @pytest.mark.asyncio
    async def test_all_methods_are_async(self, service):
        """Test that all public methods are async"""
        import inspect

        # Get all public methods (not starting with _)
        public_methods = [
            method
            for method in dir(service)
            if not method.startswith("_") and callable(getattr(service, method))
        ]

        for method_name in ["get_services", "get_gateways", "run_service"]:
            if method_name in public_methods:
                method = getattr(service, method_name)
                assert inspect.iscoroutinefunction(method)

    @pytest.mark.asyncio
    async def test_return_types(self, service, mock_client):
        """Test that methods return correct types"""
        from collections.abc import Sequence, Mapping

        # Test get_services return type
        mock_response = MagicMock()
        mock_response.json.return_value = {"result": [{"name": "test"}]}
        mock_client.get.return_value = mock_response

        services_result = await service.get_services()
        assert isinstance(services_result, Sequence)

        # Test get_gateways return type
        mock_response.json.return_value = [
            {"name": "gateway"}
        ]  # get_gateways returns direct response
        gateways_result = await service.get_gateways()
        assert isinstance(gateways_result, Sequence)

        # Test run_service return type
        mock_response.json.return_value = {"return_code": 0}
        mock_client.post.return_value = mock_response

        run_result = await service.run_service("test", "cluster")
        assert isinstance(run_result, Mapping)


class TestEdgeCases:
    """Test edge cases and boundary conditions"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    @pytest.mark.asyncio
    async def test_unicode_service_names(self, service, mock_client):
        """Test service names with Unicode characters"""
        service_name = "测试服务"
        cluster_name = "集群"
        expected_result = {"return_code": 0}

        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        await service.run_service(service_name, cluster_name)

        # Verify Unicode names were handled correctly
        expected_body = {"serviceName": service_name, "clusterId": cluster_name}
        mock_client.post.assert_called_with(
            "/gateway_manager/v1/services/run", json=expected_body
        )

    @pytest.mark.asyncio
    async def test_special_characters_in_names(self, service, mock_client):
        """Test service and cluster names with special characters"""
        service_name = "service-with-special@chars!"
        cluster_name = "cluster_with_underscores.and.dots"
        expected_result = {"return_code": 0}

        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        await service.run_service(service_name, cluster_name)

        # Should handle special characters correctly
        expected_body = {"serviceName": service_name, "clusterId": cluster_name}
        mock_client.post.assert_called_with(
            "/gateway_manager/v1/services/run", json=expected_body
        )

    @pytest.mark.asyncio
    async def test_empty_string_names(self, service, mock_client):
        """Test service and cluster names as empty strings"""
        service_name = ""
        cluster_name = ""
        expected_result = {"return_code": 1}

        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        await service.run_service(service_name, cluster_name)

        # Should handle empty strings (though may fail on server side)
        expected_body = {"serviceName": service_name, "clusterId": cluster_name}
        mock_client.post.assert_called_with(
            "/gateway_manager/v1/services/run", json=expected_body
        )

    @pytest.mark.asyncio
    async def test_very_long_names(self, service, mock_client):
        """Test very long service and cluster names"""
        service_name = "a" * 1000
        cluster_name = "b" * 1000
        expected_result = {"return_code": 0}

        mock_response = MagicMock()
        mock_response.json.return_value = expected_result
        mock_client.post.return_value = mock_response

        await service.run_service(service_name, cluster_name)

        # Should handle long names
        expected_body = {"serviceName": service_name, "clusterId": cluster_name}
        mock_client.post.assert_called_with(
            "/gateway_manager/v1/services/run", json=expected_body
        )

    @pytest.mark.asyncio
    async def test_concurrent_operations(self, service, mock_client):
        """Test concurrent service operations"""
        import asyncio

        # Setup different responses for different endpoints
        def mock_get_response(*args, **kwargs):
            endpoint = args[0]
            response = MagicMock()
            if "services" in endpoint:
                response.json.return_value = {"result": [{"name": "service1"}]}
            else:  # gateways
                response.json.return_value = [{"name": "gateway1"}]
            return response

        def mock_post_response(*args, **kwargs):
            response = MagicMock()
            response.json.return_value = {"return_code": 0}
            return response

        mock_client.get.side_effect = mock_get_response
        mock_client.post.side_effect = mock_post_response

        # Run multiple operations concurrently
        tasks = [
            service.get_services(),
            service.get_gateways(),
            service.run_service("test-service", "test-cluster"),
        ]

        results = await asyncio.gather(*tasks, return_exceptions=True)

        # All should complete successfully
        assert len(results) == 3
        for result in results:
            assert not isinstance(result, Exception)


class TestDocumentation:
    """Test service documentation and metadata"""

    @pytest.fixture
    def mock_client(self):
        """Create a mock client for testing"""
        client = AsyncMock()
        return client

    @pytest.fixture
    def service(self, mock_client):
        """Create a Service instance with mocked client"""
        service = Service(mock_client)
        return service

    def test_service_has_docstring(self, service):
        """Test that the Service class has proper documentation"""
        assert service.__class__.__doc__ is not None
        assert len(service.__class__.__doc__.strip()) > 0

    def test_methods_have_docstrings(self, service):
        """Test that all methods have proper documentation"""
        methods = [
            "get_gateways",
            "run_service",
        ]  # get_services has empty docstring currently
        for method_name in methods:
            method = getattr(service, method_name)
            assert method.__doc__ is not None
            assert len(method.__doc__.strip()) > 0

    def test_docstrings_contain_required_sections(self, service):
        """Test that method docstrings contain required sections"""
        methods = [
            "get_gateways",
            "run_service",
        ]  # get_services has empty docstring currently
        for method_name in methods:
            method = getattr(service, method_name)
            docstring = method.__doc__
            assert "Args:" in docstring
            assert "Returns:" in docstring
            assert "Raises:" in docstring
